"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from tkinter import ttk
from ..style import Color
from os.path import basename
from .menu import Menu
from .text import Text

class Treeview(ttk.Treeview):
	def __init__(self, app, *args, drag=True, **kwargs):
		if not args:
			args = [app]

		ttk.Treeview.__init__(self, *args, **kwargs)

		self.app = app
		self.menu = Menu(app)
		self.hover_item = None

		self.bind('<Motion>', self._on_motion)
		self.bind('<Leave>', self._on_leave)
		self.tag_configure("hover", background=Color.HIGHLIGHT_INACTIVE)

		if drag:
			self.drag = None
			self.bind('<B1-Motion>', self._on_b1_motion)
			self.bind('<ButtonRelease-1>', self._on_button_release_1)


	def has_tag(self, item, tag):
		if not self.exists(item): return

		tags = self.item(item)["tags"]
		return tags and tag in tags

	def add_tag(self, item, tag):
		if not self.exists(item): return

		if not isinstance(tag, tuple):
			tag = (tag,)

		tags = self.item(item)["tags"]
		if isinstance(tags, list):
			for t in tag:
				if t not in tags:
					tags.append(t)
		elif not tags:
			tags = tag
		self.item(item, tags=tags)

	def remove_tag(self, item, tag):
		if not self.exists(item): return

		if not isinstance(tag, tuple):
			tag = (tag,)

		tags = self.item(item)["tags"]
		for t in tag:
			if t in tags:
				del tags[tags.index(t)]
		self.item(item, tags=tags)

	def order(self, items=None):
		if not items: return [{
			"key": i,
			"index": self.index(i),
			"filename": basename(i)
		} for i in self.get_children()]

		items.sort(key=lambda i: i["index"])

		for i in items:
			if self.exists(i["key"]):
				self.move(i["key"], "", i["index"])
			else:
				for j in items:
					if self.exists(j["key"]) and j["filename"] == basename(i["key"]):
						self.move(j["key"], "", i["index"])

	def expanded(self, names=None, children=None):
		if children == None:
			children = self.get_children()

		if not names:
			result = []
			for j in children:
				item = self.item(j)
				if item["open"]:
					result.append({
						"name": item["text"],
						"children": self.expanded(children=self.get_children(j))
					})

			return result

		for i in names:
			for j in children:
				item = self.item(j)
				if item["text"] == i["name"]:
					self.item(j, open=True)
					self.expanded(i["children"], self.get_children(j))

	def _on_motion(self, event):
		item = self.identify_row(event.y)
		if not item or self.has_tag(item, "no_hover"): return

		if self.hover_item and item != self.hover_item:
			self.remove_tag(self.hover_item, "hover")

		self.add_tag(item, "hover")
		self.hover_item = item

	def _on_leave(self, event):
		if self.hover_item:
			self.remove_tag(self.hover_item, "hover")

		self.hover_item = None

	def _on_b1_motion(self, event):
		item, x, y = self.drag if self.drag else (None,0,0)
		selection = self.selection()
		if not selection: return

		row = self.item(selection[0])
		if not selection:
			return
		elif "." in selection[0]:
			obj_path = selection[0]
		elif self.item(selection[0])["values"]:
			obj_path = self.item(selection[0])["values"][0]
		else:
			obj_path = None

		if item:
			item.move(event.x - x, event.y - y)
		elif obj_path and selection and self.winfo_containing(event.x_root, event.y_root) == self.app.node:
			try:
				name, attr = obj_path.rsplit(".", 1)
				obj = getattr(self[name], attr)
				key = self.app.objects.setobj(attr, obj, create_new=True, coord=self.app.node.get_pointer())
				item = self.app.node[key]
			except Exception as err:
				self.app.console.showtraceback(err)
				return
				# pass
		self.drag = item, event.x, event.y

		if not item:
		    row = self.identify_row(event.y)
		    if row not in self.get_children(): return

		    moveto = self.index(row)
		    for i in self.selection():
		        self.move(i, "", moveto)

	def _on_button_release_1(self, event):
		self.drag = None


class TreeEntry(Text):
	def __init__(self, parent):
		Text.__init__(self, parent, background=Color.ALT_BACKGROUND, padx=0, pady=0)
		self.parent = parent
		self.app = parent.app
		self.editing = None
		self.bind("<Key>", self._on_key)

	def edit(self, key, event):
		name = key[:key.index("<")]
		item = self.app.node[name]
		if key == name + "<value>":
			argname = None
			in_args = False
		else:
			argname = key[key.index("<arg=") + 5:key.index(">") + 1]
			in_args = argname in item.args
		self.editing = item, argname, in_args

		value = item.obj if not argname else item.args[argname] if in_args else item.kwargs[argname]
		self.parent.selection_remove(self.parent.selection())
		self.parent.remove_tag(key, "hover")

		x,y,width,height = self.parent.bbox(key)
		y_off = self.parent.winfo_y() + height/2
		x_off = 48 if in_args or not argname else 66

		if isinstance(item.obj, str):
			self.bind("<<Modify>>", lambda event: self.highlight(r"((.*))", "yellow"))
			self.insert("end", str(value if value is not None else ""))
		else:
			self.bind("<<Modify>>", lambda event: self.syntax_highlight())
			self.insert("end", str(value if value is not None else ""))

		# self.insert("end", str(value if value is not None else ""))
		self.place(x=x + x_off, y=y + y_off, anchor="w", width=width - x_off, height=height)
		self.focus()
		self.lift()

	def finish(self, cancel=False):
		item, argname, in_args = self.editing
		content = self.get("1.0", "end-1c")
		arg = None if not argname else item.args[argname] if in_args else item.kwargs[argname]

		if isinstance(arg, str):
			value = content
		elif content in self.app.node:
			value = self.app.node[content]
		elif arg:
			try:
				value = self.app.console.eval(content)
			except:
				self.app.console.showtraceback()
				self.hide()
				return

		if not argname:
			try:
				result = self.app.console.eval(content)
			except:
				result = content

			try:
				self.app.objects[item.name] = item.obj.__class__(result)
			except Exception as err:
				self.app.console.prompt.push("")
				self.app.console.showtraceback()
				self.hide()
				return
		elif argname in item.args:
			item.args[argname] = value
		else:
			item.kwargs[argname] = value

		item.resize()
		self.hide()

	def hide(self):
		self.lower()
		self.delete("1.0", "end")
		self.editing = None

	def _on_key(self, event):
		if event.keysym == "Escape":
			self.hide()
			return "break"
		elif event.keysym == "Return":
			self.finish()
			return "break"



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import pygame
import numpy as np

def instanceof(value, classes):
    if not isinstance(classes, tuple):
        classes = classes,

    if isinstance(value, tuple):
        return tuple in classes

    if value.__class__ in (np.array, np.ndarray) and list in classes:
        return True

    if np.dtype(value.__class__) == np.int64 and int in classes:
        return True

    if np.dtype(value.__class__) == np.float64 and float in classes:
        return True

    if np.dtype(value.__class__) == np.complex128 and complex in classes:
        return True

    return isinstance(value, classes)

def color(h):
	return pygame.Color(*hex_to_rgb(h))

def hex_to_rgb(h, as_dec=False):
    result = tuple(int(h[i:i+2], 16) for i in (0, 2, 4))
    if as_dec:
        return result[0]/255, result[1]/255, result[2]/255
    return result

def is_iterable(obj):
    try:
        iter(obj)
        return True
    except:
        return False


"""
Used for parsing raw text in doc strings which cannot be parsed by other methods.
"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import re

RE_TEXT = {
    "code_sample": r"( |\t){0,}(>>>|\.\.\.).*",
    "section_title": r"(---|===)",
    "code": r"\`(.*?)\`",
    "link_url": r"(http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+~]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+)",
}

def show_textfile(text, content):
    parse = TextParser(content.strip() if content else "")
    is_title = True

    for i in parse:
        text.insert("end", i.text, i.tag)

        if i.tag in ("", "section_title"):
            text.insert("end", " \n")
            if i.tag == "section_title":
                text.insert("end", " \n", "horizontal_rule")
                text.insert("end", " \n")

    ranges = text.tag_ranges("code_sample")
    for i in range(0, len(ranges), 2):
        start = ranges[i]
        stop = ranges[i+1]
        text.syntax_highlight(start, stop)

class TextParser:
    def __init__(self, content):
        self.content = content
        self.index = 0
        self.buffer = []

    def __iter__(self):
        self.index = 0
        self.count = 0
        self.buffer.clear()
        return self

    def __next__(self):
        self.count += 1
        if len(self.buffer):
            result, tag, offset = self.buffer.pop()
            self.index += offset
            return TextNode(result, tag)

        if self.index < len(self.content):
            result = ""
            tag = ""
            is_running = True
            while is_running and self.index < len(self.content):
                text = self.content[self.index:]
                endline = -1 if "\n" not in text else text.index("\n")

                if tag == "code_sample" and not text[:endline].strip():
                    return TextNode(result, tag)

                if tag and not re.findall(RE_TEXT[tag], text[:endline]):
                    is_running = False

                for i in RE_TEXT:
                    match = re.findall(RE_TEXT[i], text[:endline])
                    if match:
                        tag = i
                        if i in ("code", "link_url"):
                            if i == "code":
                                endline = text.index("`")
                                self.buffer.append((match[0], i, 1 + len(match[0])))
                                self.index += 1 + endline
                            elif i == "link_url":
                                endline = text.index("http")
                                self.buffer.append((match[0], i, 1 + len(match[0])))
                                self.index += endline - 1
                            result += text[:endline]
                            return TextNode(result)

                if tag and tag in ("section_title"):
                    if self.count == 1:
                        is_running = False
                        self.index += 1 + endline
                    else:
                        self.buffer.append((result, "section_title", 1 + endline))
                        return TextNode("\n")
                elif endline == 0:
                    self.index += 1
                    if text[:1] == "\n":
                        if not tag and result.strip():
                            result += "\n"
                        is_running = False
                elif endline == -1:
                    result += text
                    self.index = len(self.content)
                else:
                    result += text[:endline]
                    if not tag:
                        result += " "
                    self.index += 1 + endline

                if tag == "code_sample":
                    result += "\n"

            return TextNode(result, tag)
        else:
            raise StopIteration


class TextNode:
    def __init__(self, text, tag=None):
        self.text = text
        self.tag = tag

    def __repr__(self):
        return f"{str(self.tag)}: {self.text}"



import pygame, platform
from .util import color

FONT_SIZE = 12
SPACING = 128
MARGIN = int(3 * FONT_SIZE + FONT_SIZE/2)
BACKGROUND = color("272822")
PALE_BLUE = color("c7cbd1")
BLACK = color("333333")
WHITE = color("f8f8f2")
BLUE = color("60d9f1")
RADIUS = 4

if platform.system() in ("Windows", "Linux"):
    MULTIPROCESS_CONTEXT = "spawn"
    ZOOM_MODIFIER = 5
else:
    MULTIPROCESS_CONTEXT = "fork"
    ZOOM_MODIFIER = 1



"""
Math inspector was designed to make it as simple as possible to share projects online.
To share a project on social media, simply upload the project folder to a hosting platform like
github and share the link with your friends.

To save the current project, select File > Save As from the main menu.  If the project doesn't contain any
external files or folders, then it will be saved in a file with a .math
extension, otherwise, math inspector will create a new directory on your file system with the name
of the project, and will copy any external files into this new directory, and place the .math
file in the project folder.

To open a saved project, select File > Open from the main menu, navigate to the
project folder and choose the .math file

Adding files to your project
---
You can add an existing python script to a project by selecting Modules > Add File from the main menu.
There are no restrictions on the type of python programs which are compatible.

Whenever you save your changes to a file that has been added to the current project, the node editor
is updated just as if the entire file was copy/pasted into the interpreter.  This can be very convinient for
more complicated projects.

Let's create a new file to see how it works.  Select Project > New File from the main menu and
choose a name for the file, by default it will be given a .py extension.  When you create this file,
it will appear in the left hand side panel of math inspector, right now there is nothing in the file,
but as you add things the module tab of the side panel will contain a directory of all objects available
from that file.

Try creating a complex number `z` by typing `z=1+1j` into the newly created file, and then save your changes.


README
---
If your project includes a file named README.md, then when someone opens the project the
README file will be displayed in the doc browser.

"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import pickle, os, atexit, shutil, traceback, cloudpickle
from . import plot
from tkinter import messagebox, filedialog
from .util import name_ext, AUTOSAVE_PATH
from importlib import import_module

class SaveData:
	def __init__(self, app):
		self.app = app
		self.is_first_load = True
		atexit.register(self.save)
		app.protocol("WM_DELETE_WINDOW", lambda: self.save(quit_app=True)) # save autosavefile on macos when close button in the tkinter window is clicked

	def new(self, event=None, with_dialog=True, title="Math Inspector"):
		if with_dialog:
			if not messagebox.askokcancel("MathInspector",
				"Are you sure you want to start a new file?  Any unsaved data will be lost."):
				return

		plot.close()
		self.app.title(title)
		object_keys = list(self.app.objects.store.keys())
		module_keys = list(self.app.modules.store.keys())

		for i in object_keys:
			del self.app.objects[i]

		for j in module_keys:
			del self.app.modules[j]

		children = self.app.modules.get_children()
		for k in children:
			self.app.modules.delete(k)

		self.app.modules.rootfolder = None
		self.app.modules.stop_observer()

		self.app.console.clear()
		self.app.node.zoom = 1
		self.app.console.prompt.history.clear()
		self.app.horizontal_panel.sashpos(0,0)
		self.app.vertical_panel.sashpos(0,0)
		self.app.console.do_greet()

	def save(self, file=AUTOSAVE_PATH, quit_app=False):
		if file is None or file not in (AUTOSAVE_PATH, self.app.modules.rootfolder):
			file = filedialog.asksaveasfilename(defaultextension="")
			if not file: return

		files = self.app.modules.files("file")
		folders = self.app.modules.files("folder")

		is_rootfolder = file != AUTOSAVE_PATH and len(files + folders) > 0

		if file != AUTOSAVE_PATH:
			self.app.title(os.path.basename(file))

		if is_rootfolder:
			# need to clear prev files
			# check in modified
			self._copy_project_files(file)

		objects = {}
		modules = []
		file_locals = self.app.modules.locals
		for j in self.app.modules.store:
			if not self.app.modules.has_tag(j, "local"):
				modules.append({ "alias": j, "name": self.app.modules[j].__name__})
		# modules = [{ "alias": j, "name": self.app.modules[j].__name__} for j in self.app.modules.store]
		functions = {}
		methods = {}
		custom_fn = {}
		for name in self.app.objects:
			objects[name] = self.app.objects.store[name]

		itemdata = {}
		for i in self.app.node:
			item = self.app.node[i]
			itemdata[i] = {
				"name": item.name,
				"position": item.position(),
				"connection": item.connection,
				"coord": item.position(),
				"args": { j: item.args[j] for j in item.args if j != "<value>" },
				"kwargs": { k: item.kwargs[k] for k in item.kwargs },
				"opts": item.opts.store,
			}

		plot_config = plot.config()

		try:
			help_geometry = help.browser.geometry()
		except:
			help_geometry = help.geometry

		data = [{
			"name": "app_title",
			"value": self.app.title()
		},{
			"name": "zoom",
			"value": self.app.node.zoom
		},{
			"name": "geometry",
			"value": self.app.geometry() if self.app.geometry()[:3] != "1x1" else "1280x720"
		},
		{
			"name": "horizontal_panel_sash",
			"value": self.app.horizontal_panel.sashpos(0)
		},{
			"name": "vertical_panel_sash",
			"value": self.app.vertical_panel.sashpos(0)
		},
		{
			"name": "rootfolder",
			"value": self.app.modules.rootfolder
		},{
			"name": "modules",
			"value": modules
		},{
			"name": "file_locals",
			"value": file_locals
		},{
			"name": "objects",
			"value": objects
		},{
			"name": "itemdata",
			"value": itemdata
		},{
			"name": "files",
			"value": files
		},{
			"name": "folders",
			"value": self.app.modules.files("folder")
		},{
			"name": "disabled_files",
			"value": self.app.modules.files("disabled", True)
		},{
			"name": "plot_window_size",
			"value": plot_config["size"] if plot_config else None
		},{
			"name": "plot_window_position",
			"value": plot_config["window_pos"] if plot_config else None
		},{
			"name": "plot2d_options",
			"value": { i: plot.OPTIONS_2D[i] for i in ["show_grid", "show_range"] }
		},{
			"name": "plot3d_options",
			"value": { i: plot.OPTIONS_3D[i] for i in ["show_grid"] }
		},{
			"name": "plot_window_position",
			"value": os.environ["SDL_VIDEO_WINDOW_POS"] if "SDL_VIDEO_WINDOW_POS" in os.environ else None
		},{
			"name": "help_geometry",
			"value": help_geometry
		},{
			"name": "objects_order",
			"value": self.app.objects.order()
		},{
			"name": "objects_expanded",
			"value": self.app.objects.expanded()
		},{
			"name": "console_history",
			"value": self.app.console.prompt.history.cmds
		},{
			"name": "side_select",
			"value": self.app.side_view.select()
		},{
			"name": "module_order",
			"value": self.app.modules.order()
		},{
			"name": "module_expanded",
			"value": self.app.modules.expanded()
		}]

		name, ext = name_ext(file)
		if is_rootfolder:
			filepath = os.path.join(file, os.path.basename(file) + ".math")
		else:
			filepath = os.path.join(os.path.dirname(file), name + ".math")
		with open(filepath, "wb") as output:
			cloudpickle.dump(data, output)

		if quit_app:
			plot.close()
			self.app.quit()

	def on_configure(self, name, sashpos=0):
		widget = getattr(self.app, name)
		widget.sashpos(0, sashpos)
		widget.unbind("<Configure>")

	def load(self, file=AUTOSAVE_PATH, is_first_load=False):
		if file != AUTOSAVE_PATH:
			file = filedialog.askopenfilename(filetypes=[('Math Inspector Files','*.math'), ('All Files','*.*')])
			if not file:
				return

		try:
			with open(file, "rb") as i:
				data = pickle.load(i)
		except Exception:
			self.new(with_dialog=False)
			self.app.geometry("1280x720+140+100")
			self.app.horizontal_panel.bind("<Configure>",
				lambda event: self.on_configure("horizontal_panel", 0))
			self.app.vertical_panel.bind("<Configure>",
				lambda event: self.on_configure("vertical_panel", 0))
			self.new(with_dialog=False)
			return

		data = { i["name"]:i["value"] for i in data }

		self.new(with_dialog=False, title=data["app_title"])

		try:
			if "zoom" in data:
				self.app.node.zoom = data["zoom"]
			if "geometry" in data:
				self.app.geometry(data["geometry"])
			if "horizontal_panel_sash" in data:
				self.app.horizontal_panel.bind("<Configure>",
					lambda event: self.on_configure("horizontal_panel", data["horizontal_panel_sash"]))
			if "vertical_panel_sash" in data:
				self.app.vertical_panel.bind("<Configure>",
					lambda event: self.on_configure("vertical_panel", data["vertical_panel_sash"]))
			if "modules" in data:
				for i in data["modules"]:
					self.app.menu.import_module(i["name"], i["alias"])
			if "file_locals" in data:
				self.app.modules.locals = data["file_locals"]
			if "objects" in data:
				for j in data["objects"]:
					self.app.objects.setobj(j, data["objects"][j], coord=data["itemdata"][j]["coord"])
			if "itemdata" in data:
				items = data["itemdata"]
				for m in items:
					item = self.app.node[m]
					for j in items[m]["args"]:
						item.args[j] = items[m]["args"][j]
					for k in items[m]["kwargs"]:
						item.kwargs[k] = items[m]["kwargs"][k]
					for l in items[m]["opts"]:
						item.opts[l] = items[m]["opts"][l]

				for m in self.app.node:
					item = self.app.node[m]
					if items[m]["connection"]:
						name, argname = items[m]["connection"]
						other = self.app.node[name]
						if argname in other.args.store:
							other.args[argname] = item
						elif argname in other.kwargs.store:
							other.kwargs[argname] = item
			if "rootfolder" in data:
				if data["rootfolder"]:
					self.app.modules.addfolder(data["rootfolder"], is_rootfolder=True, exec_file=False)
			if "folders" in data:
				for i in data["folders"]:
					if os.path.dirname(i) != data["rootfolder"]:
						self.app.modules.addfolder(i, exec_file=False)
			if "files" in data:
				for j in data["files"]:
					if os.path.dirname(j) != data["rootfolder"]:
						self.app.modules.addfile(j, exec_file=False)
			if "disabled_files" in data:
				for k in data["disabled_files"]:
					self.app.modules.disable_file(k)
			if "side_select" in data:
				self.app.side_view.select(data["side_select"])
			if "help_geometry" in data:
				help.geometry = data["help_geometry"]
			if "plot2d_options" in data:
				if data["plot2d_options"]:
					self.app.menu.plotconfig(**data["plot2d_options"])
			if "plot3d_options" in data:
				if data["plot3d_options"]:
					self.app.menu.plotconfig(**data["plot3d_options"])
			if "plot_window_size" in data:
				if data["plot_window_size"]:
					plot.config(size=data["plot_window_size"])
			if "plot_window_position" in data:
				if data["plot_window_position"]:
					os.environ["SDL_VIDEO_WINDOW_POS"] = data["plot_window_position"]
			if "console_history" in data:
				self.app.console.prompt.history.cmds = data["console_history"]
				self.app.console.prompt.history.i = len(data["console_history"])
			if "objects_order" in data:
				self.app.objects.order(data["objects_order"])
			if "objects_expanded" in data:
				self.app.objects.expanded(data["objects_expanded"])
			if "module_order" in data:
				self.app.modules.order(data["module_order"])
			if "module_expanded" in data:
				self.app.modules.expanded(data["module_expanded"])
			if file != AUTOSAVE_PATH and self.app.modules.rootfolder and self.app.modules.exists("README"):
				help(self.app.modules.item("README")["values"][0])

			self.app.node.scale_font()
			self.app.node.scale_width()

		except:
			traceback.print_exc()

	def _copy_project_files(self, file):
		rootfolder = os.path.splitext(file)[0]
		files = self.app.modules.files("file")
		folders = self.app.modules.files("folder")
		order = self.app.modules.order()

		for i in files + folders:
			self.app.modules.delete(i)

		if not os.path.isdir(rootfolder):
			os.mkdir(rootfolder)

		self.app.modules.stop_observer()
		for i in folders:
			if os.path.dirname(i) != rootfolder:
				shutil.copytree(i, os.path.join(rootfolder, os.path.basename(i)))

		for j in files:
			if os.path.dirname(j) != rootfolder:
				shutil.copy(j, rootfolder)

		self.app.modules.addfolder(rootfolder, is_rootfolder=True)
		self.app.modules.order(order)



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import tkinter as tk
from ..style import Color

class Button(tk.Label):
	def __init__(self, parent, *args, command=None, **kwargs):
		if "background" not in kwargs:
			kwargs["background"] = Color.ALT_BACKGROUND

		if "foreground" not in kwargs:
			kwargs["foreground"] = Color.WHITE

		if "highlightbackground" not in kwargs:
			kwargs["highlightbackground"] = Color.DARK_BLACK

		if "highlightthickness" not in kwargs:
			kwargs["highlightthickness"] = 4

		tk.Label.__init__(self, parent, *args, **kwargs)

		self.is_active = False

		self.bind("<Enter>", self._on_enter)
		self.bind("<Leave>", self._on_leave)
		self.bind("<Button-1>", self._on_button_1)
		self.bind("<ButtonRelease-1>", lambda event=None: command() if self.is_active else None)

	def _on_enter(self, event):
		self.is_active = True
		self.config(foreground=Color.DARK_GREY)

	def _on_leave(self, event):
		self.is_active = False
		self.config(foreground=Color.WHITE)

	def _on_button_1(self, event):
		self.config(foreground=Color.VERY_DARK_GREY)



"""
The interpreter overrides the function __builtins__["print"] using the displayhook from the builtins module.
The builtin_print function maintains a reference to the default python print function.
"""
builtin_print = __builtins__["print"]



"""
used for displaying markdown documents in the doc browser
"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from .regex import RE_MARKDOWN, RE_DOC

def show_markdown(text, content):
    text.delete("1.0", "end")
    text.insert("1.0", content)
    for tag in RE_MARKDOWN:
        if isinstance(RE_MARKDOWN[tag], tuple):
            for i in RE_MARKDOWN[tag]:
                text.highlight(i, tag)
        else:
            text.highlight(RE_MARKDOWN[tag], tag)

    for tag in RE_DOC:
        text.highlight(RE_DOC[tag], tag)

    text.replace(r"^(---{1,})$", "horizontal_rule", "\n")
    text.replace(r"^(==={1,})$", "horizontal_rule", "\n")

    # text.replace(r"^([-\*+])", "unordered-list", "•")
    # text.replace(r"^(\t[-\*+])", "unordered-list", "\t◦")



"""
There are two ways to animate objects, either clicking on an item
in the node editor and seleting the "Animate" option, or from the
console using the `app.animate` function.

As a simple example, the following code creates a variable `x`
and animates it's value from 0 to 1 with a delay of 0.1 seconds and a
step of 0.05

>>> x=0
>>> app.animate("x", start=0, stop=1, delay=0.1, step=0.05)

The usefulness of the app.animate function is that it is fully integrated
with the plotting features, and has been designed to optimize the
performance of plots by offloading as much of the animation to the
plot window as possible.

Another reason the app.animate function is so useful is it makes possible
to include animations in code samples that you can provide in
the documentation for your own projects.

One of the most interesting animations is the power function applied
to a set of gridlines in the complex plane.  Check it out!

>>> from mathinspector.examples import complex_grid
>>> from numpy import power
>>> t=1
>>> plot(power(complex_grid(), t))

Once the plot window is up, run the animation with the command

>>> app.animate("t", start=1, stop=4, step=0.005, delay=0.01)

You can pan and zoom while an animation is playing and the performance
should be typically be over 60 fps, as long as the mouse cursor
is inside the plot window.  If during an animation the mouse cursor is not
hovering over the plot window, the frame rate can drop.  This
happens because the plot window needs to share resources with the rest
of the app when it's not in focus.

It's also possible to animate the arguments of functions

>>> from numpy import linspace
>>> app.animate("linspace", argname="start", start=0, stop=5)

When the plot window is closed, all animations are automatically
paused. To pause or unpause all currently running animations,
use the command

>>> app.animate.pause()

"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
from . import plot
from .widget.popup import Popup

DELAY = 0.01
STEP = 0.005

def hexstring(num):
    return '#' + str(hex(num))[2:].zfill(6)

class Animation:
	def __init__(self, app):
		self.app = app
		self.is_running = False
		self.stack = []
		self.cache = {}

	def __call__(self, name, **kwargs):
		item = self.app.node[name]

		val = 0 if item.is_callable else item.obj
		opts = {
			"start": val,
			"stop": val + 10,
			"delay": DELAY,
			"step": STEP,
		}

		if kwargs:
			opts.update(kwargs)
			self.start(name, opts)
		else:
			if item.is_callable:
				opts.update({
					"argname": ""
				})

			Popup(self.app, [{
				"label": i,
				"value": opts[i]
			} for i in opts],
			lambda params: self.start(name, params),
			"Animate Item", name)

	def start(self, name, params):
		item = self.app.node[name]
		if params["stop"] is None:
			print (Exception("AnimationError: Missing required parameters"))
			return

		if "argname" in params and params["argname"] is None:
			print (Exception("AnimationError: Missing required parameters"))
			return

		self.is_running = True
		if params["start"] is None:
			pass
		elif "argname" in params:
			if params["argname"] in item.args:
				item.args[params["argname"]] = params["start"]
			elif params["argname"] in item.kwargs:
				item.kwargs[params["argname"]] = params["start"]
			else:
				print (Exception("AnimationError: argname not found"))
				return
		else:
			self.app.objects[item.name] = params["start"]

		if self.app.node.output.items:
			plot.animate(params["delay"], lambda: self._plot_update(name, params))
		else:
			self.run(item.name, params)

	def run(self, name, params):
		item = self.app.node[name]
		if "argname" in params:
			if params["argname"] in item.args:
				obj = self.app.node[name].args[params["argname"]]
			else:
				obj = self.app.node[name].kwargs[params["argname"]]
				return
		else:
			obj = self.app.objects[name]

		if self.is_running and (
			(params["step"] > 0 and obj < params["stop"])
			or (params["step"] < 0 and obj > params["stop"])
		):
			if "argname" in params:
				if params["argname"] in item.args:
					item.args[params["argname"]] += params["step"]
				else:
					item.kwargs[params["argname"]] += params["step"]
			else:
				self.app.objects[name] += params["step"]
			self.app.after(int(params["delay"]*1000), lambda: self.run(name, params))
			return

		if not self.is_running:
			self.stack.append((name, params))
			return

		self.is_running = False
		if "argname" in params:
			if params["argname"] in item.args:
				item.args[params["argname"]] = params["stop"]
			else:
				item.kwargs[params["argname"]] = params["stop"]
		else:
			self.app.objects[name] = params["stop"]

	def _plot_update(self, name, params):
		item = self.app.node[name]
		if "argname" in params:
			if params["argname"] in item.args:
				obj = item.args[params["argname"]]
			else:
				obj = item.kwargs[params["argname"]]
				return
		else:
			obj = self.app.objects[name]

		if self.is_running and (
			(params["step"] > 0 and obj < params["stop"])
			or (params["step"] < 0 and obj > params["stop"])
		):
			if "argname" in params:
				if params["argname"] in item.args:
					item.args[params["argname"]] += params["step"]
				else:
					item.kwargs[params["argname"]] += params["step"]
			else:
				self.app.objects[name] += params["step"]

			self.app.update()
			return list(self.app.node.output.values.values())

		if "argname" in params:
			if params["argname"] in item.args:
				item.args[params["argname"]] = params["stop"]
			else:
				item.kwargs[params["argname"]] = params["stop"]
		else:
			self.app.objects[name] = params["stop"]
		self.app.update()
		return False

	def can_animate(self, item):
		return item.is_callable or isinstance(item.obj, (int,float))

	def pause(self):
		if self.is_running:
			self.is_running = False
		else:
			for i in self.stack:
				i[1]["start"] = None
				self.start(i[0], i[1])
			self.stack.clear()



"""
The node editor is a block coding environment that was designed to make it as
simple as possible to experiment and interact with otherwise complex programs,
without needing to write any code, or even know much about python at all.  It's
consists of an infinte size canvas with pan, zoom, and drag functionality.  To display
the node editor, select View > Show Node Editor from the main menu.

Unlike other block coding systems, the node editor is capable of
representing arbitrary python functions and objects of any kind.  This means it's
operating at an extremely high level of generality.  This high level of generality makes
the node editor a useful tool for learning about many aspects of the python ecosystem.

Control Scheme
---
Item's in the node editor can be dragged by hovering the cursor over an item,
holding down the left mouse button, and dragging.  Holding down the left
mouse button and dragging in an empty part of the editor will create a multi-select rectangle,
which can be used for moving or deleting multiple item's at the same time.  Holding down
the right mouse button and moving the cursor pans all the items.

The scroll wheel is used for zooming in and out.

Right clicking in any empty part of the editor without panning will bring up
the "Add Objects" menu, which contains all the builtin python and math helper
functions.

You can edit the value of an item, or the argument of a function, by clicking
on it in the node editor, typing in a new value, and pressing Enter.

The Options Menu
---
Right clicking on any item will bring up a list of available options and
methods.  For example, let's create an out-of-order list `a` and use the
options menu to run the `sort` method.

Run the following code sample, then right click on
the item `a` in the node editor and select the `sort` options.

>>> a = [1,3,2]

Running the sort method from the options menu is eqivilent to running the command

>>> a.sort()

Instead of returning a sorted list, the `sort` method actually changes the value of `a`,
and this change is now reflected in the displayed value of the `a` item.

If a method has more than 0 arguments, then selecting that method from the options
menu will bring up a popup.

Methods which return objects will generally result in the creation of new items
in the editor.  However, in the case of methods which return numbers, booleans, or string,
the return value will be shown in the output log section.

Connecting Wires
---
All items have output nodes, to drag a wire from the input mode of one item and connect
it to the input node of another item, hover the mouse cursor over the output node,
hold down the left mouse button, and then drag the wire out and connect it, then let go
of the mouse button.

To plot a function, you can drag the wire to the output node section of the editor on
the right hand side.

When connecting items of different classes together, the editor will try to cast the value
of the incoming item to the class of the target item, so that the target item will preseve
it's class after connecting the wires.  If this is not possible, then the target item will
have it's class changed to be the same as the incoming item.

Add Object Menu
---
If you right click in an empty part of the node editor, it will bring up the
Add Object menu.  This menu contains all the builtin binary operators, functions,
classes, and examples.  Click on an object from the menu and it will appear in
the node editor.

Right clicking on items in the node editor will bring up a menu
of available options, including viewing the documentation.

"""
from .editor import NodeEditor


"""
The modules which are currently available in the global namespace
are displayed in the left hand sidebar of the main window.  From this
tab, you can explore the available functionality of any module, which
can be very helpful for learning new parts of the python and numpy
ecosystems.

It's possible to drag and drop any object listed in the Modules tab
into the node editor.  This can be a convinient way to instantiate
new objects without typing any code.

Right clicking in an empty part of the Modules tab will bring up a
menu of available options.  When you add a file or folder, they
will be available in the global namespace just like any other module.

The difference between importing a module and adding a file, is that
when a file is added to the Modules tab, each line of that file is
executed, and any changes that are saved to that file are updated
in the node editor, as well as across all the other views.

To directly access the global object which stores the names and values
of all modules in the global namespace, use the command

>>> app.modules

See the class `ModuleTree` for a full list of available attributes.

"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import re, os, inspect, uuid, importlib.util, sys, traceback
from tkinter import filedialog
from importlib import reload
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from .util import vdict
from .util import *
from .widget import Treeview
from .style import Color, getimage
from .console.builtin_print import builtin_print

class ModuleTree(vdict, Treeview):
	"""
	In order to synchronize the available modules across views,
	math inspector stores the aliases and values of all modules in a vdict called the Module Tree.

	When a file or folder is added to the Module Tree, the `watchdog` module
	is used to monitor changes in those files and update
	the values of variables across all views.

	Execution of code from source code is done by calling self.app.console.exec(source).
	See the documentation for `app.console` for more information.
	"""
	def __init__(self, app):
		vdict.__init__(self, getitem=self.getmodule, setitem=self.setmodule, delitem=self.delete)
		Treeview.__init__(self, app)

		self.drag = None
		self.rootfolder = None
		self.handler = None
		self.observer = None
		self.watchers = {}
		self.locals = {}
		self.bind(BUTTON_RIGHT, lambda event: self.selection_set(self.identify_row(event.y)))
		self.bind(BUTTON_RELEASE_RIGHT, self._on_button_release_right)
		self.bind("<Double-Button-1>", self._on_double_button_1)

	def getmodule(self, key):
		if key not in self.store:
			for i in self.store:
				if self.store[i].__name__ == key or self.item(i)["values"][0] == key:
					return self.store[i]
		return self.store[key]

	def setmodule(self, name, module, parent=None, file=None):
		if not parent:
			if self.exists(module.__name__):
				self.delete(module.__name__)
			self.store[name] = module
			parent = self.insert("", 'end', module.__name__,
				text="      " + name,
				values=name,
				image=getimage(".py"),
				open=True,
				tags="file" if file else "module")

		if len(self.store) == 1:
			self.app.menu.setview("modules", True)

		functions = []
		builtin_fn = []
		classes = []
		constants = []
		submodules = []
		for fn in dir(module):
			try:
				attr = getattr(module, fn)
			except:
				fn = "_" + fn

			if fn[:1] != "_":
				if inspect.isclass(attr):
					classes.append(fn)
				elif inspect.isfunction(attr):
					functions.append(fn)
				elif callable(attr):
					builtin_fn.append(fn)
				elif inspect.ismodule(attr):
					if attr.__name__ not in EXCLUDED_MODULES:# + BUILTIN_PKGS + INSTALLED_PKGS:
						submodules.append((fn, attr))
				else:
					constants.append(fn)

		if builtin_fn:
			folder = self.insert(parent, "end", text="builtins")
			for i in builtin_fn:
				self.insert(folder, "end", text=i, values=module.__name__ + "." + i)

		if functions:
			folder = self.insert(parent, "end", text="functions")
			for i in functions:
				self.insert(folder, "end", text=i, values=module.__name__ + "." + i)

		if classes:
			folder = self.insert(parent, "end", text="classes")
			for j in classes:
				self.insert(folder, "end", text=j, values=module.__name__ + "." + j)

		if constants:
			folder = self.insert(parent, "end", text="objects")
			for k in constants:
				if not re.match(r"^[A-Z_]*$", k):
					self.insert(folder, "end", text=k, values=module.__name__ + "." + k)

		if submodules:
			for fn, attr in submodules:
				if not self.exists(attr.__name__):
					folder = self.insert(parent, "end", attr.__name__, text="      " + fn, values=attr.__name__, image=getimage(".py"))
					self.setmodule(attr.__name__, attr, folder)
		return False

	def delete(self, key, del_objs=True):
		if self.has_tag(key, "folder"):
			for i in self.get_children(key):
				self.delete(i)

		if self.exists(key):
			super(Treeview, self).delete(key)
		else:
			name, ext = name_ext(key)
			if self.exists(name):
				super(Treeview, self).delete(name)

		if del_objs and key in self.locals:
			for i in self.locals[key]:
				if i in self.app.objects:
					del self.app.objects[i]
			del self.locals[key]

			if key in self.watchers:
				if self.observer:
					# TODO: use a better system for unscheduling
					try:
						self.observer.unschedule(self.watchers[key])
					except:
						pass
				del self.watchers[key]

		if key in self.store:
			del self.store[key]
		return False

	def addfile(self, file=None, parent="", index="end", watch=True, is_open=True, exec_file=True):
		file = file or filedialog.askopenfilename()
		if not file: return

		name, ext = name_ext(file)
		if os.path.basename(file) in EXCLUDED_FILES or ext in EXCLUDED_EXT: return

		if self.exists(name):
			self.delete(name, del_objs=False)

		parent = self.insert(parent if (self.rootfolder and parent != self.rootfolder) else "", index, name,
			text="      " + name,
			image=getimage(ext),
			tags="local" if parent else ("local", "file"),
			open=is_open,
			values=file)

		if ext != ".py": return

		if name in sys.modules:
			del sys.modules[name]

		if watch:
			if not self.observer:
				self.observer = Observer()
				self.observer.start()

			self.watchers[name] = self.observer.schedule(FileHandler(self, file=file),
				path=os.path.dirname(file), recursive=False)

		expanded = self.expanded()
		spec = importlib.util.spec_from_file_location(name, file)
		module = importlib.util.module_from_spec(spec)
		try:
			spec.loader.exec_module(module)
		except Exception as err:
			traceback.print_exc()
			self.disable_file(parent)
			return

		if exec_file:
			for i in dir(module):
				attr = getattr(module, i)
				if not inspect.ismodule(attr) and i in self.app.objects and i not in self.locals[name]:
					self.disable_file(parent)
					print (Exception("ImportError: object with name " + i + " already exists.\n  File \"" + file + "\", line 1, in <module>"))
					builtin_print ("\a")
					return

		if name not in self.store:
			self.store[name] = module

		self.setmodule(name, module, parent, file)
		self.expanded(expanded)

		if exec_file:
			prev = self.app.objects.store.copy()
			source = open(file, "r").read()
			try:
				self.app.console.exec(source, file)
			except:
				print ("TODO - exec failed")
				return

			new = self.app.objects.store.copy()

			if name in self.locals:
				self.locals[name].update({ k: new[k] for k in set(new) - set(prev) })
			else:
				self.locals[name] = { k: new[k] for k in set(new) - set(prev) }

			objects = [i for i in dir(module)]
			keys = list(self.locals[name].keys())
			for i in keys:
				if i not in objects:
					del self.locals[name][i]
					del self.app.objects[i]

			# for i in objects:
			# 	attr = getattr(module, i)
			# 	if inspect.ismodule(attr):
			# 		self.setmodule(i, attr, name)


	def addfolder(self, dirpath=None, parent="", watch=True, is_rootfolder=False, exec_file=True):
		dirpath = dirpath or filedialog.askdirectory(initialdir=".")
		if not dirpath: return

		name = os.path.basename(dirpath)
		parent = "" if parent == self.rootfolder else parent

		if is_rootfolder:
			self.rootfolder = dirpath
		else:
			parent = self.insert(parent, "end", dirpath,
				text="      " + name,
				open=(not parent),
				image=getimage("folder"),
				tag="folder" if not parent else "",
				values=dirpath)

		items = [(i, os.path.isdir(os.path.join(dirpath, i))) for i in os.listdir(dirpath)]
		items.sort()
		for name, is_dir in items:
			if is_dir and name not in EXCLUDED_DIR:
				self.addfolder(os.path.join(dirpath, name), parent, watch=False, exec_file=exec_file)

		for name, is_dir in items:
			if not is_dir and name not in EXCLUDED_FILES:
				fullpath = os.path.join(dirpath, name)
				self.addfile(fullpath, parent, watch=False, exec_file=exec_file)

		if watch:
			if not self.observer:
				self.observer = Observer()
				self.observer.start()

			self.watchers[name] = self.observer.schedule(FileHandler(self, dirpath=dirpath),
				path=dirpath, recursive=True)

	def stop_observer(self):
		if self.observer:
			self.observer.stop()
			self.observer = None

	def files(self, tag, recursive=False, key=""):
		children = self.get_children(key)
		if recursive is True:
			result = []
			if self.has_tag(key, tag):
				result.append(key)

			for i in children:
				result.extend(self.files(tag, True, i))
			return result

		keys = [j for j in children if self.has_tag(j, tag)]
		return [self.app.modules.item(i)["values"][0] for i in keys]

	def disable_file(self, key):
		self.item(key, image=getimage("python-disabled"))
		self.add_tag(key, "disabled")
		children = self.get_children(key)
		for i in children:
			self.delete(i)

	def _on_double_button_1(self, event):
		key = self.identify_row(event.y)
		if not key: return

		value = self.item(key)["values"]
		if not value: return
		value = value[0]

		if key in self and os.path.isfile(value):
			try:
				sourcefile = inspect.getsourcefile(self[key])
			except:
				sourcefile = None

			if sourcefile:
				open_editor(sourcefile)
				return "break"

		obj = help.getobj(value)
		if obj:
			help(obj)
			return "break"

	def _on_button_release_right(self, event):
		key = self.identify_row(event.y)
		if not key:
			self.menu.set_menu(self.app.menu.project_menu)
			return self.menu.show(event)

		value = self.item(key)["values"][0]
		items = []
		obj = help.getobj(value)
		if obj is not None:
			items.append({
				"label": "View Doc",
				"command": lambda: help(value)
			})

		try:
			file = inspect.getsourcefile(obj)
		except:
			file = None

		if file:
			items.append({
				"label": "View Source Code",
				"command": lambda: open_editor(inspect.getsourcefile(obj))
			})

		# print ('v', value)
		if key in self.get_children() and value != self.rootfolder:
			items.extend([{
				"separator": None
			}, {
				"label": "Remove " + key,
				"command": lambda: self.delete(value)
			}])

		self.menu.set_menu(items)
		self.menu.show(event)


class FileHandler(FileSystemEventHandler):
	def __init__(self, modules, file=None, dirpath=None):
		FileSystemEventHandler.__init__(self)
		self.modules = modules
		self.file = file

	def on_created(self, event):
		if self.is_excluded(event): return

		dirname = os.path.dirname(event.src_path)
		parent = "" if dirname == self.modules.rootfolder else os.path.basename(dirname)

		if event.is_directory:
			self.modules.addfolder(event.src_path, parent)
		else:
			self.modules.addfile(event.src_path, parent)

	def on_deleted(self, event):
		if self.is_excluded(event): return
		name, ext = name_ext(event.src_path)

		if event.is_directory:
			del self.modules[os.path.basename(event.src_path)]
		else:
			del self.modules[name]

	def on_modified(self, event):
		if self.is_excluded(event): return

		name, ext = name_ext(event.src_path)
		order = self.modules.order()
		expanded = self.modules.expanded()
		index = self.modules.index(name)

		dirname = os.path.dirname(event.src_path)
		parent = "" if (dirname == self.modules.rootfolder or name in self.modules) else dirname
		if parent:
			self.modules.addfile(event.src_path, parent, index=index, watch=False)
		else:
			self.modules.addfile(event.src_path, index=index, watch=False)
		self.modules.order(order)
		self.modules.expanded(expanded)

	def on_moved(self, event):
		name, ext = name_ext(event.src_path)
		self.modules.item(name, text="      " + os.path.basename(event.dest_path), values=event.dest_path)

	def is_excluded(self, event):
		name, ext = name_ext(event.src_path)
		return (self.file and self.file != event.src_path
			or event.is_directory
			or (name + ext) in EXCLUDED_FILES
			or ext in EXCLUDED_EXT
			or name in EXCLUDED_DIR
		)



from os import path
from OpenGL.GL import GL_VERTEX_SHADER, GL_FRAGMENT_SHADER, glGetUniformLocation
from OpenGL.GL import shaders

BASEPATH = path.dirname(path.abspath(__file__))

class Shader:
	def __init__(self, vert_path, frag_path, uniform=()):
		self.vert_shader = open(path.join(BASEPATH, vert_path)).read()
		self.frag_shader = open(path.join(BASEPATH, frag_path)).read()
		vert = shaders.compileShader(self.vert_shader, GL_VERTEX_SHADER)
		frag = shaders.compileShader(self.frag_shader, GL_FRAGMENT_SHADER)
		self.program = shaders.compileProgram(vert, frag)

		uniform = uniform if isinstance(uniform, tuple) else (uniform,)
		self.uniform = { i: glGetUniformLocation(self.program, i) for i in uniform }

	def __getitem__(self, key):
		return self.uniform[key]


"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .color import Color
from ..util.config import FONT

TAGS = {
    "red": {
        "foreground": Color.RED
    },

    "blue": {
        "foreground": Color.BLUE
    },

    "blue_italic": {
        "foreground": Color.BLUE,
        "font": FONT + " italic"
    },

    "green": {
        "foreground": Color.GREEN
    },

    "underline": {
        "font": FONT + " italic",
        "underline": True
    },

    "purple": {
        "foreground": Color.PURPLE
    },

    "orange": {
        "foreground": Color.ORANGE
    },

    "dark_orange_bold": {
        "foreground": Color.DARK_ORANGE,
        "font": FONT + " bold",
    },

    "console_prompt": {
        "foreground": Color.DARK_ORANGE,
        "font": FONT + " bold",
        "selectbackground": Color.DARK_BLACK,
    },

    "orange_italic": {
        "foreground": Color.ORANGE,
        "font": FONT + " italic"
    },

    "comment": {
        "foreground": Color.VERY_DARK_GREY
    },

    "yellow": {
        "foreground": Color.YELLOW
    },

    "error_file": {
        "foreground": Color.YELLOW
    },

    "error_file_hover": {
        "foreground": Color.WHITE
    },
}

TREE_TAGS = {
    "blue": {
        "foreground": Color.BLUE
    },

    "yellow": {
        "foreground": Color.YELLOW
    },

    "grey": {
        "foreground": Color.GREY
    },

    "purple": {
        "foreground": Color.PURPLE
    },

    "red": {
        "foreground": Color.RED
    },
}



"""
Usage:
    MathInspector/pkg$ python -m pip install ../ PyInstaller
    MathInspector/pkg$ python build.py
"""
import shutil
from PyInstaller.__main__ import run as pyinstaller
from platform import system

SYSTEM = system()
SPEC = {
   "Linux": "mathinspector.linux.spec",
   "Darwin": "mathinspector.macos.spec",
   "Windows": "mathinspector.win.spec",
}

shutil.rmtree("build/", ignore_errors=True)
shutil.rmtree("dist/", ignore_errors=True)
pyinstaller([SPEC[SYSTEM], "--noconfirm"])



"""
How to use the doc browser
---
In the left hand side panel you will find a directory of all available documentation
for the object you are currently viewing.  In this case, you are viewing the documentation
for the python module `doc.manual`, which is part of the math inspector source code.

Most modules and classes in python come with a lot of documentation for all their
functionality.  When you select an object from the side panel, the doc browser will display the
documentation for that object.  If you select a class or module, the side panel
will also be updated with a directory of the contents of that class or module.  To return to a
previous document, click the name of the previous object in the top navigation area.

To see how it works, try clicking on the class GettingStarted in the side panel,
and use the top navigation area to return to this document.

Running code samples
---
If you see a code sample in a document you want to run, simply click on the
code block and it will be executed by the interpreter.  Try it out!

>>> import numpy as np

Many of the code samples assume that you have already imported the numpy
package with the alias "np", as shown in the command above.  Numpy is the fundamental
package for scientific computing with python, and math inspector is designed to be a
graphical interface for interacting with the numpy family of python modules.
You can learn more about numpy at https://numpy.org

The module tree
---
If you ran the previous code example, then the module `np` should now be displayed
in the left hand side panel of the main window.  Each time a new
module is imported, the directory of it's contents will appear in the sidebar.

Math inspector comes with numpy and scipy pre-installed, as well as a number of other modules
which can be imported in a similar manner.  Any python module on your system can be
imported as long as it's compatible with python version 3.9.1

The node editor
---
The node editor is a block coding environment, and it's one of the main features of math inspector.
To display the node editor, select View > Show Node Editor from
the main menu.  It's possible to drag and drop objects, functions, and classes into the node editor directly
from the module tree.

In what follows we will use the numpy function `linspace` as an example.  From the sidebar in the main window,
expand the functions folder in the `np` module, find the function named `linspace`, and
drag it into the node editor.

An alternative way to create a `linspace` object is by using the following command

>>> from numpy import linspace

Right clicking on items in the node editor will bring up a menu of the available options.
If you right click on the `linspace` item, you will have various options, including
"show kwargs", kwargs stands for "key word arguments", and these are optional parameters
which can be passed to a function.

The `linspace` function outputs an array of evenly spaced numbers. The required arguments
are `start` and `stop`, and it has an optional keyword argument `num`, which determines
the size of the array. If you would like to learn more about linspace before continuing,
right click on the item and choose View Doc.

Displaying Plots
---
If you checked out the documentation for linspace, you may have noticed at the end there
was a code example which used matplotlib to display a plot of the output.  Math inspector's
plotting library updates and modernize the functionality available in matplotlib, and
was designed to be a replacement for matplotlib.

To plot 25 evenly spaced real numbers between -10 and 10, use the command

>>> plot(linspace(-10, 10, num=25))

To pan the camera, hold down the right mouse button and move the cursor.  You can also
use the mouse scroll wheel to zoom in and out.

Every time you update an object which is being plotted, the changes will be reflected in the plot
window. Try changing the `start` argument to the linspace item by clicking on the argument value in the
node editor.  This will bring up the text entry widget which you can use to type in a new value.
It's also possible to change the value of an argument by hovering the mouse cursor over the vaule,
holding down the left mouse button, and then dragging the cursor around.

See the modules `plot` and `animation` for more information.

Using the examples
---
Math inspector comes with a number of examples to demonstrate some of the available
functionality.  Let's import the examples module, and view the 3D plot for the cylinder
example

>>> from mathinspector import examples
>>> cylinder = examples.cylinder
>>> plot(cylinder())

You can find more information about the available examples from the documentation,
each one has a detailed explaination and multiple code samples

See the module `examples` for more information.

Further Documentation
---
Math inspector provides a number of views for interacting with the running program.
In the left hand side panel are links to the core modules which constitute
the math inspector source code.  These links contain documentation
detailing the available features, and an overview about how the code works.


Free open source software
---
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see https://www.gnu.org/licenses/.

"""

from .. import console, plot, node, objects, modules, examples, animation, project

class GettingStarted:
	"""To learn about the various features provided by math inspector, lets create
a simple program directly in the interpreter.  Click this code block to create
a variable `x` in memory with a value of `1`

>>> x=1

Math inspector provides a visual represenation of the variables in the local namespace in the
form of a block coding environment, called the node editor.  The node editor makes it possible to
experiment and interact with otherwise complex programs without needing to write any code, or
even know much about python at all.

Item's in the node editor can be dragged by hovering the cursor over an item,
holding down the left mouse button, and dragging.  Holding down
the right mouse button in an empty part of the node editor and moving the
cursor pans all the items, and the scroll wheel is used for zooming in and out.

Let's define a function `f(x)` which returns `x + 1`

>>> def f(x):
... 	return x + 1

When you call a function from the interpreter, the value of it's arguments are updated in the node editor,
as you can see by calling `f(1)`.

>>> f(1)
2

Let's now create a new variable `y` and assign it the value `f(x)`, using the variable `x` we previously created

>>> y = f(x)
>>> y
2

The wires between `x` and `f` should now be connected in the node editor, reflecting
the code we just executed.

Binary Operators
---
A binary operator is an operator like `+`,`-`,`*`,`/` which acts on the objects
on it's left and right hand sides.  To illustrate this, lets create
two new variables `a`,`b` and set them both equal to `1`

>>> a = b = 1

>>> c = a + b

When using a binary operator in a function call, a new object will appear in the
node editor representing the operator, and the wires will be connected.

>>> z = f(x+c)

You can find a full list of python's binary operators by
right clicking in an empty part of the node editor to bring up
the Add Objects menu.
	"""
	pass



COLORS = {
	"WHITE": "#f8f8f2",
	"PURE_WHITE": "#fff",
	"NAV": "#f3f3f3",
	"LEFT_NAV": "#42433e",
	"EVEN_LIGHTER_GREY": "#f5f5f5",
	"LIGHTER_GREY": "#eeeeee",
	"LIGHT_GREY": "#ececec",
	"VERY_LIGHT_GREY": "#f5f6f7",
	"GREY": "#e6e6e6",
	"DARK_GREY": "#e2e2e2",
	"COOL_GREY": "#474842",
	"DARKER_GREY": "#989898",
	"HIGHLIGHT": "#48473d",
	"HIGHLIGHT_INACTIVE": "#383830",
	"QUALITY_GREY": "#90918b",
	"VERY_DARK_GREY": "#75715d",
	"FADED_GREY": "#42433e",
	"BACKGROUND": "#272822",
	"CONSOLE_BACKGROUND": "#191919",
	"ALT_BACKGROUND": "#252526",
	"RED": "#fc1e70",
	"GREEN": "#a4e405",
	"DARK_ORANGE": "#c65d09",
	"ORANGE": "#ff9800",
	"PROMPT": "#c65d09",
	"LIGHT_PURPLE": "#3b3b62",
	"VERY_LIGHT_PURPLE": "#343460",
	"DARK_PURPLE": "#7a52b9",
	"WIRE_INACTIVE": "#7a52b9",
	"PURPLE": "#af7dff",
	"INACTIVE": "#af7dff",
	"BLUE": "#60d9f1",
	"LINK_URL": "#0088cc",
	"LINK_URL_HOVER": "#005580",
	"ACTIVE": "#60d9f1",
	"SELECT_HOVER": "#60d9f1",
	"LIGHT_BLUE": "#308bb5",
	"HOVER": "#308bb5",
	"ACTIVE_WIRE": "#308bb5",
	"WIRE_ACTIVE": "#308bb5",
	"SELECTED": "#308bb5",
	"PALE_BLUE": "#c7cbd1",
	"LIGHT_BLACK": "#555555",
	"DARK_BLACK": "#000000",
	"EMPTY_NODE": "#000",
	"BLACK": "#333333",
	"SILVER": "#bcc6cc",
	"YELLOW": "#e6dc6d",
	"LIGHT_YELLOW": "#ffffcc"
}

class ColorClass:
	def __init__(self):
		for i in COLORS:
			setattr(self, i, COLORS[i])

	def __repr__(self):
		result = {}
		for i in dir(self):
			if i[:2] != '__':
				result[i] = getattr(self, i)
		return str(result)

Color = ColorClass()



"""
The math inspector console comes with a built-in function `plot` which is
not available in a standard python environment.  As a simple example, let's
plot the point (1,1) in a two-dimensional plane.

>>> plot((1,1))

You can hold down the right mouse button and move the cursor to pan around, and use the mouse
scroll wheel to zoom in and out.  Various options are available in the Plot section of the main
menu.

To draw a line in three dimensional space from the origin to the point (10,10,10), use the command

>>> plot([(0,0,0), (10,10,10)])

Holding down the left mouse button and moving the mouse will rotate the camera,
and the WASD or arrow keys can be used for movement.  The mouse wheel is used for
scaling the entire scene.

Parameter data structure
---
`plot` accepts an arbitrary number of arguments, and each argument will be plotted at the
same time, as long as all of the arguments are either two or three dimensional.

Arguments which are integers, floats, complex numbers, or 2-tuples will be plotted
as two dimensional points, lists of these types of objects will be plotted as lines
connecting each of the specified points.  A list of a list of these types of objects
will be plotted as distinct lines.

An example of a function which draws multiple lines is complex grid

>>> from mathinspector.examples import complex_grid
>>> plot(complex_grid())

Lists of lists of 3-tuples are plotted as solid surfaces, but
you have to be careful how these lists are ordered to ensure all normal
vectors are computed properly for the lighting.  To plot multiple distinct lines
in three dimensions, pass each list of 3-tuples to `plot` as its own argument.

It's also possible to pass a function to plot which returns an array of pixels,
by using the pixelmap key word argument.

An example of a pixelmap function is the Domain Coloring function

>>> from mathinspector.examples import domain_coloring
>>> plot(pixelmap=domain_coloring)

Animating pixelmaps is very computationally expensive, and the
`multiprocess` module is used to try and do as much of the computation in the background
as possible.  The plot window creates an offscreen surface for the purpose
of panning and zooming, and after movement stops for 1 second, it recalculates
the pixels for the current position, while also recomputing the pixels for the
offscreen surface in the background.  This system currently suffers from some flashing and
resolution artifacts which are hard to miss when moving very quickly.

Look for video rendering functionality in an upcoming release to address the
performance issues associated with the most computationally expensive animations.
"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from os import environ
environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "1"
import pygame

from .plot2d import SDLWindow, OPTIONS as OPTIONS_2D
from .plot3d import OpenGLWindow, OPTIONS as OPTIONS_3D
from .util import instanceof

active_window = None
opengl_window = OpenGLWindow()
sdl_window = SDLWindow()

def plot(*args, **kwargs):
	global active_window

	window = get_window(*args, **kwargs)

	if not window:
		raise Exception("Invalid Parameters")
	elif is_active() and window != active_window:
		print(Exception("Can't plot 2D and 3D at the same time."))
		return
	active_window = window
	active_window.plot(*args, **kwargs)

def is_active():
	return pygame.display.get_init()

def get_window(*args, **kwargs):
	if "pixelmap" in kwargs:
		return sdl_window

	for arg in args:
		if callable(arg):
			return False
		if instanceof(arg, (tuple,list)) and len(arg) == 3 and instanceof(arg[0], list) and instanceof(arg[0][0], (int,float)):
			return opengl_window
		if instanceof(arg, (int,float,complex)):
			return sdl_window if len(args) > 1 else False
		if instanceof(arg, tuple):
			if len(arg) == 1 and instanceof(arg[0], complex):
				return sdl_window
			if len(arg) == 2 and instanceof(arg[0], (int,float)):
				return sdl_window
			if len(arg) == 3 and instanceof(arg[0], (int,float)):
				return opengl_window
			else:
				return get_window(*arg)
		if instanceof(arg, list):
			if len(arg) == 3 and instanceof(arg[0], list) and len(arg[0]) == len(arg[1]) == len(arg[2]):
				return opengl_window
			return get_window(*arg)
	return False

def config(**kwargs):
	if kwargs:
		OPTIONS_2D.update(kwargs)
		OPTIONS_3D.update(kwargs)
		if active_window and pygame.get_init():
			active_window.config(**kwargs)
		return

	if not active_window:
		return OPTIONS_2D
	return OPTIONS_2D if active_window == sdl_window else OPTIONS_3D

def animate(delay, callback):
	active_window.animate(delay, callback)

def update(*args, **kwargs):
	if active_window:
		active_window.update(*args, **kwargs)
		return

def close():
	if pygame.display.get_init():
		active_window.close()



"""
Used for preprocessing and post-processing commands strings in order to perform any updates to the
various views throughout the app which are not captured by other methods.
"""
"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import ast
from ..util import binop

class CodeParser(ast.NodeVisitor):
	def __init__(self, app):
		ast.NodeVisitor.__init__(self)
		self.app = app
		self.call_visitor = CallVisitor(app)
		self.assign_visitor = AssignVisitor(app)

	def preprocess(self, source):
		try:
			tree = ast.parse(source)
		except:
			return

		self.call_visitor.visit(tree)
		return tree

	def postprocess(self, source):
		try:
			tree = ast.parse(source)
		except:
			return

		## NOTE: the reason for removing the assign visitor is to ensure the math inspector interpreter
		##			always exec's all sequences of command strings, and produces the same output,
		##			exactly as would occur when using the python command line interpreter
		# self.assign_visitor.visit(tree)  ## TODO - add this as a toggle option in the main menu, default is turned off
		return tree

class CallVisitor(ast.NodeVisitor):
	def __init__(self, app):
		ast.NodeVisitor.__init__(self)
		self.app = app

	def visit_Call(self, node):
		if isinstance(node.func, ast.Name) and node.func.id in self.app.node:
			item = self.app.node[node.func.id]
			for i, val in enumerate(node.args):
				if item:
					if len(item.argspec[0]) > i:
						argname = item.argspec[0][i]
						attr = item.args
					elif len(item.argspec[1]) > 0:
						argname = list(item.kwargs.store.items())[i - len(item.argspec[0])][0]
						attr = item.kwargs
						item.opts["show_kwargs"] = True
					else:
						return

				if isinstance(val, ast.Name):
					if val.id in self.app.node:
						attr[argname] = self.app.node[val.id]
				elif isinstance(val, ast.Num):
					attr[argname] = val.n
				elif isinstance(val, ast.Str):
					attr[argname] = val.s
				elif (isinstance(val, ast.BinOp)
					and isinstance(val.left, ast.Name)
					and isinstance(val.right, ast.Name)
					and val.left.id in self.app.node
					and val.right.id in self.app.node
				):
					fn = get_binop(val.op)
					name = fn + "_" + val.left.id + val.right.id
					self.app.objects.setobj(name, getattr(binop, fn))
					_binop = self.app.node[name]
					_binop.args["a"] = self.app.node[val.left.id]
					_binop.args["b"] = self.app.node[val.right.id]
					attr[argname] = _binop
				elif isinstance(val, ast.Call) and val.func.id in self.app.node:
					self.visit_Call(val)
					attr[argname] = self.app.node[val.func.id]
				elif isinstance(val, ast.UnaryOp):
					if isinstance(val.op, ast.USub) and isinstance(val.operand, ast.Num):
						attr[argname] = -val.operand.n
			for k in node.keywords:
				if isinstance(k.value, ast.Name):
					if k.value.id in self.app.node:
						item.kwargs[k.arg] = self.app.node[k.value.id]
						item.opts["show_kwargs"] = True
				elif isinstance(k.value, ast.Num):
					item.kwargs[k.arg] = k.value.n
					item.opts["show_kwargs"] = True
				elif isinstance(k.value, ast.Str):
					item.kwargs[k.arg] = k.value.s
					item.opts["show_kwargs"] = True
		elif isinstance(node.func, ast.Name) and node.func.id == "plot":
			for val in node.args:
				if isinstance(val, ast.Name):
					if val.id in self.app.node:
						self.app.node.output.connect(self.app.node[val.id], show_plot=False)
				elif isinstance(val, ast.Call) and val.func.id in self.app.node:
					self.visit_Call(val)
					self.app.node.output.connect(self.app.node[val.func.id], show_plot=False)
		elif isinstance(node.func, ast.Attribute):
			if isinstance(node.func.value, ast.Name) and node.func.value.id in self.app.node:
				self.app.node[node.func.value.id].value(self.app.objects[node.func.value.id])

class AssignVisitor(ast.NodeVisitor):
	def __init__(self, app):
		ast.NodeVisitor.__init__(self)
		self.app = app

	def visit_Assign(self, node):
		target = node.targets[0]
		val = node.value
		if isinstance(target, ast.Name) and target.id in self.app.node:
			item = self.app.node[target.id]
			if isinstance(val, ast.Call):
				if isinstance(val.func, ast.Name) and val.func.id in self.app.node:
					if val.func.id in self.app.node:
						if "<value>" in item.args:
							item.args["<value>"] = self.app.node[val.func.id]
						else:
							self.app.objects[val.func.id] = item.value()
			elif (isinstance(val, ast.BinOp)
				and isinstance(val.left, ast.Name)
				and isinstance(val.right, ast.Name)
				and val.left.id in self.app.node
				and val.right.id in self.app.node
			):
				fn = get_binop(val.op)
				name = fn + "_" + val.left.id + val.right.id
				self.app.objects.setobj(name, getattr(binop, fn))
				_binop = self.app.node[name]
				_binop.args["a"] = self.app.node[val.left.id]
				_binop.args["b"] = self.app.node[val.right.id]
				item.args["<value>"] = _binop

def get_binop(node):
	for i in dir(binop):
		if isinstance(node, getattr(ast, i)):
			return i



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
# import tkinter as tk
from tkinter import ttk
from ttkthemes import themed_tk
# import multiprocessing as mp
from .console import Interpreter
from .node import NodeEditor
from .objects import ObjectTree
from .modules import ModuleTree
from .mainmenu import MainMenu
from .project import SaveData
from .animation import Animation
from .widget import Notebook


class App(themed_tk.ThemedTk):
	def __init__(self):
		themed_tk.ThemedTk.__init__(self)
		self.set_theme("arc")
		ttk.Style(self)

		self.horizontal_panel = ttk.PanedWindow(self, orient="horizontal")
		self.vertical_panel = ttk.PanedWindow(self, orient="vertical")
		self.side_view = Notebook(self, has_labels=True)

		self.node = NodeEditor(self)
		self.console = Interpreter(self)
		self.modules = ModuleTree(self)
		self.objects = ObjectTree(self)

		self.vertical_panel.add(self.node.frame)
		self.vertical_panel.add(self.console.frame)
		self.side_view.add("Modules", self.modules)
		self.side_view.add("Objects", self.objects)
		self.horizontal_panel.add(self.side_view.frame)
		self.horizontal_panel.add(self.vertical_panel)
		self.horizontal_panel.pack(side="left", fill="both", expand=True)

		self.animate = Animation(self)
		self.project = SaveData(self)
		self.menu = MainMenu(self)
		self.config(menu=self.menu)
		self.project.load(is_first_load=True)


def main():
	app = App()
	app.mainloop()



"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import pygame, os, platform
import numpy as np
import multiprocessing as mp
from .config import *
from .util import instanceof, is_iterable
from pygame.locals import *
from pygame._sdl2.video import Window

OPTIONS = {
	"title": "Math Inspector",
	"show_grid": True,
	"show_range": False,
	"resolution": 1,
	"size": (1280, 720),
	"position": None,
	"step": 1/128,
	"timeout": 1,
	"on_update": None,
	"on_close": None,
	"pixelmap": None,
	"window_pos": None,
	"fontsize": 12
}

class SDLWindow:
	def __init__(self):
		self.zoom = 1
		self.screen = None
		self.font = None
		self.is_animation_running = False
		self.scale = 1
		self.args = []
		self.spacing = 1 / OPTIONS["step"]
		if platform.system() not in ("Windows", "Linux"):
			self.ctx = mp.get_context(MULTIPROCESS_CONTEXT)
			self.queue = self.ctx.Queue()

	def get_pixels(self, res, size):
		x0, y0 = OPTIONS["position"]
		w, h = OPTIONS["size"]
		step = OPTIONS["step"]
		return OPTIONS["pixelmap"](((-x0+w/2)*step, (y0-h/2)*step), (w*size*step,h*size*step), res*step)

	def draw_pixels(self):
		w,h = OPTIONS["size"]
		pix = self.get_pixels(1,1)
		pygame.surfarray.blit_array(self.screen, pix[0:w,0:h])
		if OPTIONS["show_grid"]:
			self.draw_grid()
		for values in self.args:
			self.draw_values(values)
		if OPTIONS["show_range"]:
			self.draw_range()
		pygame.display.flip()

	def update_offscreen(self, res=1, size=3):
		if self.is_processing or platform.system() in ("Windows", "Linux"): return
		self.is_processing = True
		w,h = OPTIONS["size"]
		self.coords.append([w/self.pan_res,h/self.pan_res,res])
		process = self.ctx.Process(target=self.process_target, args=(res,size))
		process.start()

	def process_target(self, res, size):
		self.queue.put(self.get_pixels(res,size))

	def plot(self, *args, **kwargs):
		self.args = args
		OPTIONS.update(kwargs)
		w, h = OPTIONS["size"]

		pygame.init()
		pygame.display.set_caption(OPTIONS["title"])
		self.screen = pygame.display.set_mode((w,h), pygame.RESIZABLE)

		# TODO - get position from OPTIONS
		if not OPTIONS["position"]:
			OPTIONS["position"] = w/2, h/2
		x0, y0 = OPTIONS["position"]

		is_running = True
		is_focused = True
		did_change = True
		offscreen_surf = None
		self.pan_res = 1
		pixels = None
		self.coords = [[w/3, h/3, self.pan_res]]
		mode = None
		self.is_processing = False
		timer = 0
		animation_timer = 0
		request_quit = False
		last_tick = 0
		clock = pygame.time.Clock()

		if OPTIONS["on_update"]:
			OPTIONS["on_update"]()

		if OPTIONS["pixelmap"]:
			offscreen_surf = pygame.Surface((0,0))
			offscreen_size = offscreen_surf.get_size()
			try:
				self.draw_pixels()
			except Exception as err:
				if OPTIONS["on_close"]:
					OPTIONS["on_close"]()
				pygame.display.quit()
				print (err)
				return
			self.update_offscreen()
			did_change = False

		while is_running:
			tick = pygame.time.get_ticks()/1000
			delta_time = tick - last_tick
			last_tick = tick

			keypress = pygame.key.get_pressed()
			for event in pygame.event.get(pump=False):
				if event.type == pygame.ACTIVEEVENT:
					is_focused = bool(event.gain)
					# is_focused = not bool(event.state)
				elif event.type == pygame.VIDEORESIZE:
					did_change = True
					OPTIONS["size"] = w, h = pygame.display.get_surface().get_size()
				elif event.type == pygame.MOUSEMOTION:
					if event.buttons[2]:
						did_change = True
						mode = "pan"
						x0 += event.rel[0]
						y0 += event.rel[1]
						for i, val in enumerate(self.coords):
							self.coords[i][0] -= event.rel[0]/val[2]
							self.coords[i][1] -= event.rel[1]/val[2]
				elif event.type == pygame.MOUSEWHEEL:
					did_change = True
					mode = "zoomin" if event.y > 0 else "zoomout"
					delta = event.y*ZOOM_MODIFIER/SPACING
					self.zoom *= 1 + delta
					x, y = pygame.mouse.get_pos()
					x0 += delta*(x0 - x)
					y0 += delta*(y0 - y)

					for i, val in enumerate(self.coords):
						self.coords[i][2] *= 1 + delta
						self.coords[i][0] += x*delta/val[2]
						self.coords[i][1] += y*delta/val[2]

					if self.scale * self.zoom > 1:
						self.scale /= 2
					elif self.scale * self.zoom < 1/2:
						self.scale *= 2
					self.spacing = self.scale * self.zoom * SPACING
				elif event.type == pygame.USEREVENT:
					if hasattr(event, "args") and event.args:
						self.args = event.args
					elif hasattr(event, "kwargs"):
						OPTIONS.update(event.kwargs)
						if "pixelmap" in kwargs:
							self.draw_pixels()
							offscreen_surf.fill(BACKGROUND)
							self.update_offscreen()
					elif hasattr(event, "animate"):
						delay = event.animate[0]
						callback = event.animate[1]
						self.is_animation_running = True
					did_change = True
				elif (event.type == pygame.QUIT
					or (event.type == KEYUP and event.key == K_ESCAPE)
					or (event.type == KEYUP and event.key == K_w and keypress[K_LMETA])
					or (event.type == KEYUP and event.key == K_q and keypress[K_LMETA])
				):
					request_quit = True

			if OPTIONS["on_update"] and not is_focused:
				OPTIONS["on_update"]()

			if self.is_animation_running:
				if animation_timer >= delay:
					animation_timer = 0
					values = callback()
					if values:
						self.args = values
						did_change = True
					else:
						self.is_animation_running = False
				else:
					animation_timer += delta_time

			if platform.system() not in ("Windows", "Linux"):
				if not self.queue.empty():
					pixels = self.queue.get()

					if offscreen_size != (pixels.shape[0], pixels.shape[1]):
						offscreen_surf = pygame.Surface((pixels.shape[0], pixels.shape[1]))
						offscreen_size = offscreen_surf.get_size()

					pygame.surfarray.blit_array(offscreen_surf, pixels)
					self.coords.pop(0)
					self.is_processing = False

			if request_quit and not self.is_processing:
				is_running = False

			# if keypress[K_w] or keypress[K_UP]:
			# 	y0 += 5
			# 	did_change = True
			# if keypress[K_s] or keypress[K_DOWN]:
			# 	y0 -= 5
			# 	did_change = True
			# if keypress[K_d] or keypress[K_RIGHT]:
			# 	x0 -= 5
			# 	did_change = True
			# if keypress[K_a] or keypress[K_LEFT]:
			# 	x0 += 5
			# 	did_change = True

			if did_change:
				timer = 0
				OPTIONS["position"] = x0, y0
				OPTIONS["step"] = step = self.scale / self.spacing
				if OPTIONS["pixelmap"] is not None and platform.system() not in ("Windows", "Linux"):
					self.screen.fill(BACKGROUND)
					x1, y1, zoom2 = self.coords[0]
					if len(self.coords) == 1 and not self.is_processing and (
						zoom2 > 2
						or zoom2 < 0.5
						or not w/(2*zoom2) < x1 < pixels.shape[0] - w/(2*zoom2)
						or not h/(2*zoom2) < y1 < pixels.shape[1] - h/(2*zoom2)
					):
						self.update_offscreen()

					if 0 < x1 < offscreen_size[0] - w/zoom2 and 0 < y1 < offscreen_size[1] - h/zoom2:
						s1 = offscreen_surf.subsurface((x1,y1),(w/zoom2, h/zoom2))
						pygame.transform.scale(s1, (w,h), self.screen)
					else:
						self.screen.fill(BACKGROUND)

				else:
					self.screen.fill(BACKGROUND)

				if OPTIONS["show_grid"]:
					self.draw_grid()

				for values in self.args:
					self.draw_values(values)

				if OPTIONS["show_range"]:
					self.draw_range()
				pygame.display.flip()
				did_change = False
			elif OPTIONS["pixelmap"] and 0 <= timer < OPTIONS["timeout"]:
				timer += delta_time

				if timer >= OPTIONS["timeout"]:
					self.draw_pixels()

			if not request_quit:
				pygame.event.pump()

		self.is_animation_running = False
		win_w, win_h = Window.from_display_module().position
		OPTIONS["window_pos"] = str(win_w) + ", " + str(win_h)
		os.environ["SDL_VIDEO_WINDOW_POS"] = OPTIONS["window_pos"]
		pygame.display.quit()

		if OPTIONS["on_close"]:
			OPTIONS["on_close"]()

	def draw_values(self, obj):
		if instanceof(obj, (int, float, complex)):
			pygame.draw.circle(self.screen, BLUE, self.get_points(obj), RADIUS)
		elif instanceof(obj, tuple):
			if len(obj) == 2 and instanceof(obj[0], (int, float, complex)):
				pygame.draw.circle(self.screen, BLUE, self.get_points(obj), RADIUS)
			else:
				for i in obj:
					self.draw_values(i)
		elif instanceof(obj, list):
			if instanceof(obj[0], complex):
				pygame.draw.aalines(self.screen, BLUE, False, self.get_points(obj))
			elif instanceof(obj[0], list) and len(obj[0]) == 2:
				pygame.draw.aalines(self.screen, BLUE, False, self.get_points(obj))
			elif instanceof(obj[0], tuple):
				pygame.draw.aalines(self.screen, BLUE, False, self.get_points(obj))
			else:
				for i in obj:
					self.draw_values(i)

	def draw_grid(self):
		x0, y0 = OPTIONS["position"]
		w, h = OPTIONS["size"]

		# sub grid lines
		s1 = pygame.Surface((w,h), SRCALPHA)
		alpha = int(255*(2 * self.spacing/SPACING - 1))
		_black = pygame.Color(51,51,51,alpha)
		x = x0 % self.spacing + self.spacing/2
		while x < w:
			pygame.draw.line(s1, _black, [x, 0], [x, h], 1)
			x += self.spacing

		y = y0 % self.spacing + self.spacing/2
		while y < h:
			pygame.draw.line(s1, _black, [0, y], [w, y], 1)
			y += self.spacing

		self.screen.blit(s1, (0,0))

		# main grid lines
		x = x0 % self.spacing
		while x < w:
			pygame.draw.line(self.screen, BLACK, [x, 0], [x, h], 1)
			x += self.spacing

		y = y0 % self.spacing
		while y < h:
			pygame.draw.line(self.screen, BLACK, [0, y], [w, y], 1)
			y += self.spacing

		# axis
		pygame.draw.line(self.screen, PALE_BLUE, [x0, 0], [x0, h], 1)
		pygame.draw.line(self.screen, PALE_BLUE, [0, y0], [w, y0], 1)

		# origin marker
		pygame.draw.line(self.screen, WHITE, [x0 - 4, y0], [x0 + 4,y0], 3)
		pygame.draw.line(self.screen, WHITE, [x0, y0 - 4], [x0,y0 + 4], 3)

	def draw_range(self):
		x0, y0 = OPTIONS["position"]
		w, h = OPTIONS["size"]

		if not self.font:
			self.font = pygame.font.SysFont("Monospace", 12)

		# margin
		pygame.draw.rect(self.screen, BLACK, (0,h-MARGIN, w, h))
		pygame.draw.rect(self.screen, BLACK, (0,0, MARGIN, h))

		# range number
		x = self.spacing + x0 % self.spacing
		while x < w - MARGIN:
			x_mark = str(round((x - x0) * self.scale/self.spacing, 2))
			text = self.font.render(x_mark, False, PALE_BLUE)
			self.screen.blit(text,(x - 4*len(x_mark), h - (OPTIONS["fontsize"] + MARGIN)/2))
			x += self.spacing

		y = y0 % self.spacing
		while y < h - MARGIN:
			y_mark = str(-round((y - y0) * self.scale/self.spacing, 2))
			text = self.font.render(y_mark, False, PALE_BLUE)
			self.screen.blit(text,(round(MARGIN / (1 + len(y_mark))), y - 8))
			y += self.spacing

	def get_points(self, points):
		x0, y0 = OPTIONS["position"]
		step = OPTIONS["step"]
		if instanceof(points, list):
			if len(points) == 2 and instanceof(points[0], (int, float)):
				return x0 + points[0]/step, y0 - points[1]/step
			return [self.get_points(p) for p in points]
		if instanceof(points, (int, float)):
			return x0 + points/step, y0
		if instanceof(points, complex):
			return x0 + points.real/step, y0 - points.imag/step
		if instanceof(points, tuple):
			return x0 + points[0]/step, y0 - points[1]/step

	def update(self, *args, **kwargs):
		if args:
			pygame.event.post(pygame.event.Event(pygame.USEREVENT, args=args))

	def config(self, **kwargs):
		pygame.event.post(pygame.event.Event(pygame.USEREVENT, kwargs=kwargs))

	def animate(self, delay, callback):
		pygame.event.post(pygame.event.Event(pygame.USEREVENT, animate=(delay,callback)))

	def close(self):
		pygame.event.post(pygame.event.Event(pygame.QUIT))



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import tkinter as tk
from tkinter import ttk
from ..style import TAGS, Color
from ..util.config import BUTTON_RIGHT, BUILTIN_FUNCTION, BUILTIN_CLASS
from ..widget.menu import Menu
from ..console.builtin_print import builtin_print
import io, re, tokenize, keyword, builtins, inspect

RE_PY = {

	"orange_italic": r"^def (\w+)\s?\(((\w+(,\s?)?)+)\)",
	#"orange_italic": r"(\w+)\s?\(((\w+(,\s?)?)+)\)",
	# "orange_italic": r".([a-zA-Z0-9_]*)=(.*?)( |\n)*(,|\))",
	"green": r"(def {1,}(\w+))",
	"blue": r"((\w+)\()",
	"purple": r"((True|False|None))",
	"console_prompt": r"((>>>|\.\.\.))"
}

# REFACTOR - this is a good pattern, move this into style
DEFAULT_OPTS = {
	"foreground": Color.WHITE,
	"background": Color.BACKGROUND,
	"font": "Monospace 14",
	"insertbackground": Color.WHITE,
	"padx": 16,
	"pady": 8,
	"borderwidth": 0,
	"highlightthickness": 0,
	"selectbackground": Color.HIGHLIGHT,
	"inactiveselectbackground": Color.HIGHLIGHT_INACTIVE,
	"tabs": ("3c","4c", "5c"),
	"undo": True,
	"wrap": "word"
}


class Text(tk.Text):
	def __init__(self, parent, *args, readonly=False, has_scrollbar=False, **kwargs):
		opts = DEFAULT_OPTS.copy()
		opts.update(kwargs)
		tk.Text.__init__(self, parent, *args, **opts)

		if has_scrollbar:
			self.scrollbar = tk.Scrollbar(parent, command=self.yview)
			self.config(yscrollcommand=self.scrollbar.set)
			self.scrollbar.config(command=self.yview)
			# self.scrollbar.pack(side="right", fill="y")
			# self.bind("<<Scroll>>", self._on_scroll)

		for i in TAGS:
			self.tag_configure(i, **TAGS[i])

		if readonly:
			self.bind("<Key>", self._on_key)

		self.readonly = readonly
		self.hover_range = None

		self._orig = self._w + "_orig"
		self.tk.call("rename", self._w, self._orig)
		self.tk.createcommand(self._w, self._proxy)
		self.bind(BUTTON_RIGHT, self._on_button_right)
		self.menu = Menu(self)

	def _on_key(self, event):
		ctrl = (event.state & 0x4) != 0
		meta = (event.state & 0x8) != 0
		is_mod = ctrl or meta
		return None if is_mod and event.char == "c" else False if is_mod else "break"

	def insert(self, *args, syntax_highlight=False, **kwargs):
		idx = self.index(self.index(args[0]) + "-1c")
		linestart = self.index(idx + " linestart")
		super(Text, self).insert(*args, **kwargs)
		if syntax_highlight:
			self.syntax_highlight(linestart, idx + "+" + str(len(args[1])) + "c")


	def replace(self, pattern, tag=None, newtext=None, start="1.0", end="end"):
		current_index = "1.0"
		while True:
			content = self.get(current_index, end)
			match = re.compile(pattern, re.MULTILINE).search(content)
			if not match:
				break
			start_index=match.start()
			end_index=match.end()
			ind1 = current_index + "+" + str(start_index) + "c"
			ind2 = current_index + "+" + str(end_index) + "c"
			self.delete(ind1, ind2)
			replacement = newtext if newtext != None else match.group(1)
			# if pattern[-1] == "$":
			# 	replacement += "\n"
			self.insert(ind1, replacement, tag)
			current_index = self.index(ind1 + "+" + str(len(replacement)) + "c")


	def highlight(self, pattern, tag, start="1.0", end="end"):
		for match in re.compile(pattern, re.MULTILINE).finditer(self.get(start, end)):
			ind1 = self.index(self.index(start) + "+" + str(match.start(2)) + "c")
			ind2 = self.index(self.index(start) + "+" + str(match.end(2)) + "c")
			self.tag_add(tag, ind1, ind2)

	def syntax_highlight(self, start="1.0", end="end"):
		for i in TAGS:
			self.tag_remove(i, start, end)

		content = self.get(start, end)
		if not content: return

		try:
			for i in RE_PY:
				self.highlight(RE_PY[i], i, start, self.index(end))

			for typ, string, start_index, end_index, line in tokenize.generate_tokens(io.StringIO(content).readline):
				token = tokenize.tok_name[typ]

				ind1 = self.index(self.index(start) + "+" + str(start_index[0] - 1) + "l" + "+" + str(start_index[1]) + "c")
				ind2 = self.index(self.index(start) + "+" + str(end_index[0] - 1) + "l" + "+" + str(end_index[1]) + "c")

				if token == "NAME":
					if string in ["def", "clear", "app", "help", "plot", "class"]:
						self.tag_add("blue_italic", ind1, ind2)
					elif string in keyword.kwlist:
						self.tag_add("red", ind1, ind2)
					elif string in BUILTIN_FUNCTION:
						self.tag_add("blue", ind1, ind2)
					elif string in BUILTIN_CLASS:
						self.tag_add("blue_italic", ind1, ind2)
				elif token == "OP":
					if string in ["=", "*", "**"]:
						self.tag_add("red", ind1, ind2)
				elif token == "STRING":
					self.tag_add("yellow", ind1, ind2)
				elif token == "NUMBER":
					self.tag_add("purple", ind1, ind2)
				elif token == "COMMENT":
					self.tag_add("comment", ind1, ind2)
		except Exception as err:
			pass

	def remove_selection(self):
		tag_ranges = self.tag_ranges("sel")
		if tag_ranges:
			self.tag_remove("sel", *tag_ranges)

	def _proxy(self, *args):
		try:
			cmd = (self._orig,) + args
			result = self.tk.call(cmd)
		except:
			return

		if args[0] in ("insert", "replace", "delete"):
			self.event_generate("<<Modify>>")

		if args[0:2] == ("yview", "scroll"):
			self.event_generate("<<Scroll>>")

		if args[0:3] == ("mark", "set", "insert"):
			self.event_generate("<<MarkSet>>")

		return result

	def _on_button_right(self, event):
		tag_ranges = self.tag_ranges("sel")
		if self.readonly:
			if tag_ranges:
				self.menu.show(event, [{
					"label": "Copy",
					"command": lambda: self.clipboard_append(self.get(*tag_ranges))
				}])
			return

		if tag_ranges:
			self.menu.show(event, [{
				"label": "Copy",
				"command": lambda: self.clipboard_append(self.get(*tag_ranges))
			}, {
				"label": "Paste",
				"command": lambda: self._on_paste(tag_ranges)
			}])
		else:
			self.menu.show(event, [{
				"label": "Paste",
				"command": lambda: self.insert("insert", self.clipboard_get())
			}])

	def _motion(self, event, tag):
		self.config(cursor="pointinghand")
		hover_range = self.tag_prevrange(tag, "@" + str(event.x) + "," + str(event.y))
		if not hover_range: return

		if self.hover_range and hover_range != self.hover_range:
			self.tag_remove(tag + "_hover", *self.hover_range)

		self.hover_range = hover_range
		content = self.get(*hover_range)
		if content:
			# print ("ddd", tag + "_hover", content)
			self.tag_add(tag + "_hover", *hover_range)

	def _leave(self, event, tag):
		self.config(cursor="")
		if self.hover_range:
			self.tag_remove(tag + "_hover", *self.hover_range)

	def get_tags(self, start, end):
	    index = start
	    tags = []
	    while self.compare(index, "<=", end):
	        tags.extend(self.tag_names(index))
	        index = self.index(f"{index}+1c")

	    return tuple(set(tags))



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from collections import OrderedDict

class vdict(dict): # need to extend dict so it can be passed as an argument to exec()
	"""
	vdict: a dictionary class that has callbacks for get, set, and del events
	"""
	def __init__(self, *args, getitem=None, setitem=None, delitem=None, **kwargs):
		self.store = OrderedDict()
		self.store.update(dict(*args, **kwargs))
		self._get = getitem
		self._del = delitem
		self._set = setitem
		self.keys = self.store.keys
		self.values = self.store.values

	def __getitem__(self, key):
		if self._get:
			return self._get(key)
		return self.store[key]

	def __setitem__(self, key, value):
		if self._set:
			ret = self._set(key, value)
			if ret is False:
				return
		self.store[key] = value

	def __delitem__(self, key):
		if self._del:
			ret = self._del(key)
			if ret is False:
				return
		del self.store[key]

	def __contains__(self, key):
		return key in self.store

	def __iter__(self):
		return iter(self.store)

	def __len__(self):
		return len(self.store)

	def __repr__(self):
		return repr(self.store)


"""
used for loading images from ../assets
"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os, sys
from PIL import ImageTk, Image
from ..util.config import BASEPATH

ASSET_PATH = os.path.join(BASEPATH, "img")

IMAGE = {
	".py": "pythonlogo-color.png",
	"python-disabled": "pythonlogo-disabled.png",
	".md": "markdown.png",
	"folder": "folder.png",
	"folder-open": "folder-open.png",
	"blank-file": "blank-file.png",
	"node": "node.png",
	"node-connect": "node-connect.png",
	"node-active": "node-active.png",
	"function-f": "function-f.png",
}

cache = {}

def init():
	for i in IMAGE:
		cache[i] = ImageTk.PhotoImage(Image.open(os.path.join(ASSET_PATH, IMAGE[i])))

def getimage(key):
	if len(cache) == 0:
		init()
	if key in cache:
		return cache[key]
	return cache["blank-file"]



"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import math, builtins
from .. import examples
from ..util import BUILTIN_FUNCTION, BUILTIN_CLASS, BUILTIN_CONSTANT, binop

TRIG_FUNCTIONS = [i for i in ("acos", "acosh", "asin", "asinh", "atan", "atan2", "atanh", "cos", "cosh", "degrees", "sin", "sinh", "tan", "tanh")]
MATH_FUNCTIONS = [i for i in dir(math) if callable(getattr(math, i)) and i not in TRIG_FUNCTIONS and i[:1] != "_"]
MATH_CONSTANTS = [i for i in dir(math) if i not in MATH_FUNCTIONS + TRIG_FUNCTIONS and i[:1] != "_"]

def object_menu(app):
	return [{
		"label": "Operator",
		"menu": [{
			"label": i,
			"command": lambda key=i: create_object(app, key, binop)
		} for i in dir(binop) if i[0] != "_"]
	},{
		"label": "Math",
		"menu": [{
			"label": "Constant",
			"menu":[{
				"label": i,
				"command": lambda key=i: create_object(app, key, np if hasattr(np, key) else math)
			} for i in MATH_CONSTANTS]
		},{
			"separator": None
		},{
			"label": "Trig",
			"menu":[{
				"label": i,
				"command": lambda key=i: create_object(app, key, np if hasattr(np, key) else math)
			} for i in TRIG_FUNCTIONS]
		},{
			"separator": None
		}] + [{
			"label": i,
			"command": lambda key=i: create_object(app, key, np if hasattr(np, key) else math)
		} for i in MATH_FUNCTIONS]
	},{
		"label": "Builtin Class",
		"menu": [{
			"label": str(i),
			"command": lambda key=i: create_object(app, key, builtins)
		} for i in BUILTIN_CLASS if i[:1] != "_" and not i[:1].isupper() and "Error" not in i and "Warning" not in i]
	},{
		"label": "Builtin Function",
		"menu": [{
			"label": str(i),
			"command": lambda key=i: create_object(app, key, module=builtins)
		} for i in BUILTIN_FUNCTION if i[:1] != "_" and not i[:1].isupper()]
	},{
		"label": "Examples",
		"menu": [{
			"label": str(i).replace("_", " ").capitalize(),
			"command": lambda key=i: create_object(app, key, module=examples)
		} for i in dir(examples) if i[:1] != "_" and not i[:1].isupper() and i != "np"]
	}]

def create_object(app, name, module):
	obj = getattr(module, name)
	app.objects.setobj(name, obj, create_new=True)



"""
class which manages tab autocomplete and integrates with the rest of the app.
"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import builtins, re
from functools import reduce
from ..util import BUILTIN_PKGS, BUILTIN_FUNCTION, BUILTIN_CLASS, INSTALLED_PKGS, KEYWORD_LIST
from .builtin_print import builtin_print

RE_WORD = re.compile(r"[() ]")

class AutoComplete():
	def __init__(self, prompt):
		self.prompt = prompt
		self.console = prompt.console
		self.tabcount = 0

	def __call__(self):
		content = self.prompt.get().strip()
		self.tabcount += 1
		s = RE_WORD.split(content)[-1] #REFACTOR - split to also split by any nonword characters (besides .) like paren
		query = s.rsplit(".", 1)
		is_import = (content[:7] == "import ")
		if is_import:
			choices = [i for i in self.console.app.modules] + BUILTIN_PKGS + INSTALLED_PKGS
		elif s.count("."):
			choices = self.console.eval("dir(" + query[0] + ") + " + "dir(" + str(query[0].__class__.__name__) + ")")
		else:
			self.console.synclocals()
			choices = [i for i in self.console.locals] + BUILTIN_FUNCTION + BUILTIN_CLASS + KEYWORD_LIST

		if not choices:
			builtin_print("\a")
			return

		result = []
		for j in choices:
			word = content[7:] if is_import else query[0] if len(query) == 1 else query[1]
			if j[:len(word)] == word:
				result.append(j)

		if not result:
			builtin_print("\a")
			return

		common = findcommonstart(result)

		if len(result) == 1:
			self.prompt.insert("end", result[0][len(word):])
			self.tabcount = 0
		elif word[:len(common)] != common:
			self.prompt.insert("end", common[len(word):])
		elif self.tabcount > 0:
			self.console.write("\n" + "        ".join(result) + "\n\n")
			self.prompt.move()
			self.tabcount = 0
		else:
			builtin_print("\a")
			self.tabcount += 1

def getcommonletters(strlist):
    return ''.join([x[0] for x in zip(*strlist) if reduce(lambda a,b:(a == b) and a or None,x)])

def findcommonstart(strlist):
    strlist = strlist[:]
    prev = None
    while True:
        common = getcommonletters(strlist)
        if common == prev:
            break
        strlist.append(common)
        prev = common

    return getcommonletters(strlist)


"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import inspect, os, platform
from ..style import Color

def getrandom(min, max):
    N = 10
    max_val, min_val = 1, -1
    range_size = (max_val - min_val)  # 2
    np.random.rand(N) * range_size + min_val

def fontcolor(value, as_string=False):
    result = "blue"
    if isinstance(value, str):
        result = "yellow"
    elif isinstance(value, dict):
        result = "grey"
    elif isinstance(value, int) or isinstance(value, float) or isinstance(value, complex):
        result = "purple"
    elif callable(value) or value is None:
        result = "red"
    return result if as_string else getattr(Color, result.upper())

def classname(obj):
    if inspect.isclass(obj):
        return obj.__name__
    elif inspect.isfunction(obj) and obj.__module__:
        return obj.__module__ + "." + obj.__name__
    elif obj.__class__.__module__ and hasattr(obj, "__name__"):
        return obj.__class__.__module__ + "." + obj.__name__
    return obj.__class__.__name__

def name_ext(file):
    return os.path.splitext(os.path.basename(file))

def open_editor(file):
    if platform.system() in ("Windows", "Linux"):
        command = "start " + file
    elif platform.system() == "Darwin":
        command = "open " + file
    os.system(command)

def instanceof(value, classes):
    if not isinstance(classes, tuple):
        classes = classes,

    if isinstance(value, tuple) and tuple in classes:
        return True

    if np.dtype(value.__class__) == np.object_ and list in classes:
        return True

    if np.dtype(value.__class__) == np.int64 and int in classes:
        return True

    if np.dtype(value.__class__) == np.float64 and float in classes:
        return True

    if np.dtype(value.__class__) == np.complex128 and complex in classes:
        return True

    return isinstance(value, classes)



"""
Leverages the parsing by docutils FunctionDoc in numpy, which is used for generating the numpy
documentation, and generates the tkinter text tags for each parsed section.
"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from .regex import RE_DOC

def show_functiondoc(text, doc, name):
    text.insert("1.0", name + "\n", ("h1", "name"))
    for i in doc:
        if len(doc[i]) > 0:
            text.insert("end", "\n")
            if i == "Signature":
                signature = doc[i].replace(", /", "").replace(", \*", "").replace(", *", "")
                text.insert("end", signature, "signature")
                text.syntax_highlight("end-1c linestart")
                text.insert("end", "\n")
            elif i  == "Summary":
                text.insert("end", "".join(doc[i]) + "\n")
            elif i  == "Extended Summary":
                text.insert("end", strjoin(doc[i]).replace("\n", "\n\n") + "\n")
            elif i  in ("Other Parameters", "Attributes", "Methods", "Warnings"):
                text.insert("end", i + "\n", "see_also_title")
                text.insert("end", " ".join(doc[i]) + "\n")
            elif i in ("Parameters", "Returns", "Raises", "Yields", "Warns"):
                text.insert("end", i.upper() + "\n", ("section_title", "bordered_box", "spacing_top"))
                for j in doc[i]:
                    text.insert("end", j[0], ("parameter_name", "bordered_box"))
                    text.insert("end", " : ", ("bordered_box"))
                    text.insert("end", j[1], ("italic", "blue", "bordered_box"))
                    text.insert("end", "\n", ("bordered_box"))
                    text.insert("end", strjoin(j[2]), ("parameter_description", "bordered_box", "spacing_bottom"))
                    text.insert("end", "\n", ("bordered_box"))
            elif i == "See Also":
                text.insert("end", "See also:\n", ("see_also", "see_also_title"))
                for j in doc[i]:
                    text.insert("end", j[0], ("doc_link", "see_also"))
                    if len(j[1]) > 0:
                        text.insert("end", " : " + "".join(j[1]) + "\n", "see_also")
                    else:
                        text.insert("end", "\n", "see_also")
                text.insert("end", "\n", "see_also")
            elif i == "Notes":
                text.insert("end", "Notes\n", "underline_title")
                text.insert("end", "\n", "horizontal_rule")
                text.insert("end", "\n", "")
                text.insert("end", strjoin(doc[i]) + "\n", "notes")
            elif i == "References":
                items = [j.strip() for j in doc[i]]
                text.insert("end", "References\n", "underline_title")
                text.insert("end", "\n", "horizontal_rule")
                text.insert("end", "\n")
                text.insert("end", " ".join(items).replace(".. ", "\n\n") + "\n", "references")
            elif i == "Examples":
                text.insert("end", "Examples\n", "underline_title")
                text.insert("end", "\n", "horizontal_rule")
                text.insert("end", "\n")
                text.insert("end", "\n".join(doc[i]) + "\n")

    for tag in RE_DOC:
        text.highlight(RE_DOC[tag], tag)

    ranges = text.tag_ranges("code_sample")
    for i in range(0, len(ranges), 2):
        start = ranges[i]
        stop = ranges[i+1]
        text.syntax_highlight(start, stop)

def strjoin(items):
    return "".join([i + " " if len(i) > 0 else "\n" for i in items])


"""
Math Inspector: a visual programming environment for scientific computing
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import tkinter as tk
import pygame
from ..widget import Text
from ..util import argspec, instanceof, MESSAGE_TIMEOUT
from ..console.builtin_print import builtin_print
from ..style import Color, getimage
from .. import plot

class Output(tk.Frame):
	def __init__(self, canvas):
		self.canvas = canvas
		self.app = canvas.app
		tk.Frame.__init__(self, canvas.frame, background=Color.BLACK)
		self.log = Text(canvas.frame,
			font="Nunito 24 bold",
			foreground=Color.ORANGE,
			background=Color.BLACK,
			readonly=True,
			bd=1
		)

		plot.config(on_update=self.app.update, on_close=self.disconnect)

		self.node = tk.Label(self, image=getimage("node"), foreground=Color.RED, padx=0, pady=0, bd=0)
		self.node.pack()
		self.node.bind("<Enter>", self._on_enter_node)
		self.node.bind("<Leave>", self._on_leave_node)
		self.node.bind("<Configure>", self._on_configure)
		self.log.tag_configure("center", justify="center")
		self.items = []
		self.values = {}
		self.log_item = None

	def show(self, obj, autohide=True):
		self.log.delete("1.0", "end")
		self.log.insert("end", str(obj), "center")
		self.log.place(x=0, y=0, width=self.canvas.winfo_width(), height=60)
		self.log.lift()

		if autohide:
			self.after(MESSAGE_TIMEOUT, self.log.lower)

	def _on_enter_node(self, event):
		if len(self.items) == 1:
			self.hover()
			self.items[0].config("output", "wire", fill=Color.HOVER)

	def _on_leave_node(self, event):
		if self.items:
			self.hover()
			if len(self.items) == 1:
				self.items[0].config("output", "wire", fill=Color.PURPLE)

	def hover(self, is_visible=True):
		if is_visible is False:
			img = "node-active" if self.items else "node"
		else:
			img = "node-connect"
		self.node.config(image=getimage(img))

	def connect(self, item, show_plot=True):

		value = item.value()
		window = plot.get_window(value)
		if not window and not show_plot:
			return

		if show_plot and window and plot.is_active() and window != plot.active_window:
			item.move_wire()
			print(Exception("Can't plot 2D and 3D at the same time."))
			return

		self.items.append(item)
		item.move_wire()
		self.hover()
		if not show_plot: return

		name = " ".join([i.name for i in self.items])

		if self.is_pixelmap(item.obj):
			pixelmap_fn = lambda *args: item.obj(*args, **item.kwargs)
			if plot.is_active():
				plot.config(pixelmap=pixelmap_fn)
			else:
				plot.plot(pixelmap=pixelmap_fn, title=name)
		elif window:
			self.values[item.name] = value
			if plot.is_active():
				plot.update(*list(self.values.values()))
			else:
				plot.plot(value, title=name)
		else:
			if self.log_item:
				self.disconnect(self.log_item)
			self.show(item.content(truncate=False), autohide=False)
			self.log_item = item

	def disconnect(self, item=None):
		if item is None:
			for i in self.items:
				i.hide_wire()
				i.config("output", fill=Color.EMPTY_NODE)
			self.items.clear()
			self.values.clear()
		else:
			self.items = [i for i in self.items if i != item]
			if self.log_item == item:
				self.log_item = None
			elif item.name in self.values:
				del self.values[item.name]
			item.hide_wire()
			item.config("output", fill=Color.EMPTY_NODE)

		if not self.items:
			self.hover(False)
			self.log.lower()
			if plot.is_active() and item:
				plot.close()
		else:
			plot.update(*list(self.values.values()))

	def is_pixelmap(self, obj):
		if not callable(obj):
			return False
		return argspec(obj)[0] == ["position", "size", "step"]

	def update_value(self, item):
		if plot.is_active():
			if self.is_pixelmap(item.obj):
				plot.config(pixelmap=lambda *args: item.obj(*args, **item.kwargs))
			else:
				self.values[item.name] = item.value()
				plot.update(*list(self.values.values()))
		else:
			self.show(item.content(truncate=False), autohide=False)

	def _on_configure(self, event):
		for i in self.items:
			i.move_wire()

		if self.items and not plot.is_active():
			self.log.place(width=self.canvas.winfo_width())



"""
The objects which are currently in the local namespace are displayed in the left hand sidebar
of the main window, in the "Objects" tab.  The Objects tab is a traditional object
debugger, which you might be familiar with from web inspector and related
development tools in modern web browsers.

From this tab, you can change the values of objects, or the value of arguments
and key word arguments to any function, simply by clicking on the relevant item
and typing in a new value.

A list of an object's methods are contained in the methods folder.  To run
a method, either double click on it, or right click and choose "Run Method"
from the options menu.  If the method doesn't have any arguments, it will
be run, otherwise the method dialog will appear. It's also possible to drag
methods directly into the node editor from the Objects tab.

To directly access the global object which stores the names and values
of all variables, use the command

>>> app.objects

See the class `ObjectTree` for a full list of available attributes.
"""
"""
Copyright (C) 2021 Matt Calhoun

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import inspect
from . import doc
from .util import vdict, Color, argspec, classname, fontcolor, open_editor, BUTTON_RIGHT, BUTTON_RELEASE_RIGHT
from .style import TREE_TAGS
from .widget import Treeview, TreeEntry

class ObjectTree(vdict, Treeview):
	"""
	In order to synchronize the values of variables currently in memory across views,
	math inspector stores the names and values of all variables in a vdict called the Object Tree.
	A vdict is like a normal python dictionary, except it has callbacks for things like setting and
	deleting items.

	A vdict stores its values in an ordered dictionary attribute named "store".
	To display the current contents of the Object Tree, use the command

	>>> app.objects.store

	As an example, if you wanted to set the variable `x` to have a value of `1`, you
	could use the command

	>>> x=1

	which is equivilent to

	>>> app.objects["x"] = 1

	if you need fine grained control when assigning variables to the local namespace,
	use the function

	>>> app.objects.setobj("x", 1)

	See the documentation for `setobj` for more information
	about the available keyword arguments
	"""
	def __init__(self, app, *args, **kwargs):
		vdict.__init__(self, *args, setitem=self.setobj, delitem=self.delete)
		Treeview.__init__(self, app, *args, **kwargs)

		self.drag = None
		self.entry = TreeEntry(self)

		for i in TREE_TAGS:
			self.tag_configure(i, **TREE_TAGS[i])

		self.tag_configure("class", foreground=Color.BLUE, font="Nunito 10")
		self.tag_configure("editable")
		self.tag_configure("argname", foreground=Color.DARK_ORANGE, font="Nunito 10")
		self.tag_configure("kwargname", foreground=Color.ORANGE, font="Nunito 10")

		self.bind("<Button-1>", self._on_button_1)
		self.bind("<Double-Button-1>", self._on_double_button_1)
		self.bind(BUTTON_RIGHT, self._on_button_right)
		self.bind(BUTTON_RELEASE_RIGHT, self._on_button_release_right)

	def setobj(self, name, value, create_new=False, coord=None, is_binop=False):
		"""
		The callback of app.objects, called whenever new objects are created, or the value
		of an object is changed.  This method is responsible for updating the values
		displayed in the Objects tab of the sidebar, as well as notifying the other
		views that the value of a variable in the local namespace has changed.

		Parameters
		----------
		name : string
		    the name of the variable
		value : object
		    the value to assign
		create_new : bool
			when set to True, this will generate a unique variable name if there is already
			a key with the same name
		coord : tuple(float, float)
		    a tuple of floats which determines the coordinates of the newly created item in the
		    node editor


		Examples
		--------
		>>> app.objects.setitem("x", create_new=True)
		>>> app.objects.setitem("x", create_new=True)
		"""
		name = self.unique_name(name) if create_new else name
		prev = self.app.node[name] if name in self.app.node else None
		order = self.order()
		expanded = self.expanded()
		update_value = prev and prev.argspec == argspec(value) and prev.classname == classname(value)
		is_output_item = prev in self.app.node.output.items or prev == self.app.node.output.log_item
		self.store[name] = value
		if is_output_item and not update_value:
			self.app.node.output.disconnect(prev)
			is_output_item = False
		self.app.node.setitem(name, update_value=update_value, coord=coord, is_output_item=is_output_item)

		if update_value and self.exists(name + "<arg=<value>>"):
			self.item(name + "<arg=<value>>",
				text=str(value),
				tags=("editable", fontcolor(value, as_string=True)))
			return

		if self.exists(name):
			super(Treeview, self).delete(name)

		item = self.app.node[name]

		self.insert("", "end", name, text=name, tags=("object", "doc"))
		self.insert(name, "end", name + "<class>", text=item.classname, tags=("class", "no_hover"))
		if not item.is_callable:
			self.insert(name, "end", name + "<arg=<value>>", text=str(value), tags=("editable", fontcolor(value, as_string=True)))

		if item.is_callable:
			for j in item.args:
				self.insert(name, "end", name + "<argname=" + j + ">", text=j,
					tags=("argname", "no_hover"))
				self.insert(name, "end", name + "<arg=" + j + ">",
					text=str(item.args[j]),
					tags=("editable", fontcolor(item.args[j], as_string=True)))

			if len(item.kwargs) > 0:
				kwargs = self.insert(name, "end", name + "<kwargs>", text="kwargs")
				for k in item.kwargs:
					self.insert(kwargs, "end", name + "<argname=" + k + ">", text=k,
						tags=("kwargname", "no_hover"))
					self.insert(kwargs, "end", name + "<arg=" + k + ">",
						text=str(item.kwargs[k]),
						tags=("editable", fontcolor(item.kwargs[k], as_string=True)))

		for j in dir(value):
			if j[:1] != "_":
				attr = getattr(value, j)
				if callable(attr):
					if not self.exists(name + "<methods>"):
						self.insert(name, "end", name + "<methods>", text="methods")
					self.insert(name + "<methods>", "end", name + "." + j, text=j, tags="doc")
				else:
					if not self.exists(name + "<constants>"):
						self.insert(name, "end", name + "<constants>", text="constants")
					self.insert(name + "<constants>", "end", name + "." + j, text=j + ": " + str(attr), tags="no_hover")

		self.order(order)
		self.expanded(expanded)

		if create_new:
			return name
		return False

	def delete(self, key):
		if key in self.app.node:
			item = self.app.node[key]
			item.disconnect()
			for i in item.args["connection"]:
				item.args["connection"][i].disconnect()
			for k in item.kwargs["connection"]:
				item.kwargs["connection"][k].disconnect()
		super(Treeview, self).delete(key)
		self.app.node[key].destroy()

	def unique_name(self, name):
	    i = 2
	    temp = name
	    while True:
	        if temp not in self:
	            return temp
	        temp = name + "_" + str(i)
	        i += 1

	def _on_button_1(self, event):
		key = self.identify_row(event.y)
		if not key: return

		if self.entry.editing:
			self.entry.finish(cancel=True)

		if self.has_tag(key, "editable"):
			self.entry.edit(key, event)
			return "break"

		if self.has_tag(key, "no_hover"):
			return "break"

	def _on_double_button_1(self, event):
		key = self.identify_row(event.y)
		obj = help.getobj(key)
		if callable(obj) and "." in key:
			name, fn = key.rsplit(".", 1)
			self.app.node.run_method(fn, obj, self.app.node[name])

	def _on_button_right(self, event):
		key = self.identify_row(event.y)
		self.selection_set(key)

	def _on_button_release_right(self, event):
		key = self.identify_row(event.y)
		obj = help.getobj(key)
		if obj is None: return

		items = []

		if help.getobj(key):
			items.append({
				"label": "View Doc",
				"command": lambda: help(key)
			})

		try:
			file = inspect.getsourcefile(self.app.objects[key])
			items.append({
				"label": "View Source Code",
				"command": lambda: open_editor(file)
			})
		except:
			pass

		if key in self.app.node and self.app.animate.can_animate(self.app.node[key]):
			items.append({
				"label": "Animate",
				"command": lambda: self.app.animate(key)
			})

		if callable(obj) and "." in key:
			name, fn = key.rsplit(".", 1)
			items.append({
				"label": "Run Method",
				"command": lambda: self.app.node.run_method(fn, obj, self.app.node[name])
			})

		if key in self:
			items.extend([{
				"separator": None
			},{
				"label": "Delete " + key,
				"command": lambda: self.delete(key)
			}])

		self.menu.set_menu(items)
		self.menu.show(event)



"""
Maps the ast BinOp classes to functions that can be used in the node editor
"""
"""
Copyright (C) 2021 Matt Calhoun
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
def Add(a,b):
	"""
	a + b : a plus b
	"""
	return a + b

def Sub(a,b):
	"""
	a - b : a minus b
	"""
	return a - b

def Mult(a,b):
	"""
	a * b : a times b
	"""
	return a * b

def Div(a,b):
	"""
	a / b : a divided by b
	"""
	return a / b

def FloorDiv(a,b):
	"""
	a // b : the largest integer less than a divided by b
	"""
	return a // b

def Mod(a,b):
	"""
	a % b : a modulo b
	"""
	return a % b

def Pow(a,b):
	"""
	a ** b : a to the power b
	"""
	return a ** b

def LShift(a,b):
	"""
	a << b : left shift bitwise operator
	"""
	return a << b

def RShift(a,b):
	"""
	a >> b : right shift bitwise operator
	"""
	return a >> b

def BitOr(a,b):
	"""
	a | b : bitwise or operator
	"""
	return a | b

def BitXor(a,b):
	"""
	a ^ b : bitwise xor operator
	"""
	return a ^ b

def BitAnd(a,b):
	"""
	a & b : bitwise and operator
	"""
	return a & b